/*******************************************************************************
This file contains:
    definitions of the triggers between schema, objects, and effectors.
    definitions and descriptions of the schema hierarchy, internal links

*******************************************************************************/

#include "cs.h"

extern CsSchema *CsInstallSchemaNode(CsSchema **schemas, char *schema_name, char *schema_goal, int level, ActionType action);
extern void CsInstallSchemaTriggeringFunctions(CsSchema *schema, BOOLEAN (*obj)(), BOOLEAN (*res)(), double (*sit)());
extern void CsInstallSchemaSubgoal(CsSchema *schema, char *subgoal, ArgumentFunction arg_function, GoalType sort);
extern void CsInstallOrderingConstraint(CsSchema *schema, char *subgoal1, char *subgoal2);
extern void CsInstallSchemaResource(CsSchema *schema, Resource resource);

/******************************************************************************/

extern BOOLEAN breakfast_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN coffee_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN consume_cereal_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN sugar_from_pack_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN sugar_from_bowl_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN coffee_from_jar_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN coffee_from_pack_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN milk_from_carton_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN stir_with_stirrer_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN cereal_into_mouth_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN drink_from_mug_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN pick_up_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN put_down_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN dip_spoon_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN empty_spoon_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN consume_spoon_contents_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN stir_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN pour_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN tear_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN screw_open_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN screw_closed_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN swap_triggers_object(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN drink_triggers_object(CsObject *object, ArgumentFunction function, int cycle);

extern BOOLEAN breakfast_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN coffee_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN consume_cereal_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN sugar_from_pack_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN sugar_from_bowl_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN coffee_from_jar_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN coffee_from_pack_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN milk_from_carton_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN stir_with_stirrer_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN cereal_into_mouth_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN drink_from_mug_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN pick_up_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN put_down_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN dip_spoon_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN empty_spoon_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN consume_spoon_contents_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN stir_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN pour_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN tear_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN screw_open_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN screw_closed_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN swap_triggers_effector(CsResource *effector, int cycle);
extern BOOLEAN drink_triggers_effector(CsResource *effector, int cycle);

extern double situation_triggers_breakfast(int cycle);
extern double situation_triggers_coffee(int cycle);
extern double situation_triggers_consume_cereal(int cycle);
extern double situation_triggers_sugar_from_pack(int cycle);
extern double situation_triggers_sugar_from_bowl(int cycle);
extern double situation_triggers_coffee_from_jar(int cycle);
extern double situation_triggers_coffee_from_pack(int cycle);
extern double situation_triggers_milk_from_carton(int cycle);
extern double situation_triggers_stir_with_stirrer(int cycle);
extern double situation_triggers_cereal_into_mouth(int cycle);
extern double situation_triggers_drink_from_mug(int cycle);
extern double situation_triggers_pick_up(int cycle);
extern double situation_triggers_put_down(int cycle);
extern double situation_triggers_dip_spoon(int cycle);
extern double situation_triggers_empty_spoon(int cycle);
extern double situation_triggers_consume_spoon_contents(int cycle);
extern double situation_triggers_stir(int cycle);
extern double situation_triggers_pour(int cycle);
extern double situation_triggers_tear(int cycle);
extern double situation_triggers_screw_open(int cycle);
extern double situation_triggers_screw_closed(int cycle);
extern double situation_triggers_swap(int cycle);
extern double situation_triggers_drink(int cycle);

CsSchema *CsInitialiseCoffeePreparationSchemas()
{
    CsSchema *schemas = NULL;
    CsSchema *schema;

/********* Level 3 ************************************************************/

    schema = CsInstallSchemaNode(&schemas, "Prepare Instant Coffee", "prepare_coffee", 3, NONE);
    CsInstallSchemaSubgoal(schema, "add_milk", EMPTY, CRUX);
    CsInstallSchemaSubgoal(schema, "add_coffee_grinds", EMPTY, CRUX);
    CsInstallSchemaSubgoal(schema, "sweeten", EMPTY, CRUX);

    CsInstallOrderingConstraint(schema, "add_coffee_grinds", "add_milk");
    CsInstallOrderingConstraint(schema, "add_coffee_grinds", "sweeten");
/* */
    CsInstallSchemaTriggeringFunctions(schema, coffee_triggers_object, coffee_triggers_effector, situation_triggers_coffee);

    schema = CsInstallSchemaNode(&schemas, "Prepare Cereal", "consume_cereal", 3, NONE);
    CsInstallSchemaSubgoal(schema, "add_milk", EMPTY, CRUX);
    CsInstallSchemaSubgoal(schema, "sweeten", EMPTY, CRUX);
    CsInstallSchemaSubgoal(schema, "eat_cereal", EMPTY, CRUX);
    CsInstallOrderingConstraint(schema, "add_milk", "sweeten");
    CsInstallOrderingConstraint(schema, "sweeten", "eat_cereal");
    CsInstallSchemaTriggeringFunctions(schema, consume_cereal_triggers_object, consume_cereal_triggers_effector, situation_triggers_consume_cereal);

/********* Level 2 ************************************************************/

    schema = CsInstallSchemaNode(&schemas, "Sugar from Packet", "sweeten", 2, NONE);
    CsInstallSchemaSubgoal(schema, "goal_put_down", EMPTY, CLEANUP);
    CsInstallSchemaSubgoal(schema, "goal_pour", TARGET, CRUX);
    CsInstallSchemaSubgoal(schema, "goal_open", EMPTY, PRECONDITION);
    CsInstallSchemaSubgoal(schema, "goal_pick_up", SOURCE, PRECONDITION);
    CsInstallOrderingConstraint(schema, "goal_pour", "goal_put_down");
    CsInstallSchemaTriggeringFunctions(schema, sugar_from_pack_triggers_object, sugar_from_pack_triggers_effector, situation_triggers_sugar_from_pack);

    schema = CsInstallSchemaNode(&schemas, "Sugar from Bowl", "sweeten", 2, NONE);
    CsInstallSchemaSubgoal(schema, "goal_put_down", EMPTY, CLEANUP);
    CsInstallSchemaSubgoal(schema, "goal_empty_spoon", TARGET, CRUX);
    CsInstallSchemaSubgoal(schema, "goal_dip_spoon", SOURCE, PRECONDITION);
    CsInstallSchemaSubgoal(schema, "goal_pick_up", IMPLEMENT, PRECONDITION);
    CsInstallOrderingConstraint(schema, "goal_empty_spoon", "goal_put_down");
    CsInstallSchemaTriggeringFunctions(schema, sugar_from_bowl_triggers_object, sugar_from_bowl_triggers_effector, situation_triggers_sugar_from_bowl);

    schema = CsInstallSchemaNode(&schemas, "Coffee from Packet", "add_coffee_grinds", 2, NONE);
    CsInstallSchemaSubgoal(schema, "goal_put_down", EMPTY, CLEANUP);
    CsInstallSchemaSubgoal(schema, "goal_pour", TARGET, CRUX);
    CsInstallSchemaSubgoal(schema, "goal_open", EMPTY, PRECONDITION);
    CsInstallSchemaSubgoal(schema, "goal_pick_up", SOURCE, PRECONDITION);
    CsInstallOrderingConstraint(schema, "goal_pour", "goal_put_down");
    CsInstallSchemaTriggeringFunctions(schema, coffee_from_pack_triggers_object, coffee_from_pack_triggers_effector, situation_triggers_coffee_from_pack);

    schema = CsInstallSchemaNode(&schemas, "Coffee from Jar", "add_coffee_grinds", 2, NONE);
    CsInstallSchemaSubgoal(schema, "goal_put_down", EMPTY, CLEANUP);
    CsInstallSchemaSubgoal(schema, "goal_empty_spoon", TARGET, CRUX);
    CsInstallSchemaSubgoal(schema, "goal_dip_spoon", SOURCE, PRECONDITION);
    CsInstallSchemaSubgoal(schema, "goal_pick_up", IMPLEMENT, PRECONDITION);
    CsInstallOrderingConstraint(schema, "goal_empty_spoon", "goal_put_down");
    CsInstallSchemaTriggeringFunctions(schema, coffee_from_jar_triggers_object, coffee_from_jar_triggers_effector, situation_triggers_coffee_from_jar);

    schema = CsInstallSchemaNode(&schemas, "Milk from Carton", "add_milk", 2, NONE);
    CsInstallSchemaSubgoal(schema, "goal_put_down", EMPTY, CLEANUP);
    CsInstallSchemaSubgoal(schema, "goal_pour", TARGET, CRUX);
    CsInstallSchemaSubgoal(schema, "goal_open", ANY, PRECONDITION);
    CsInstallSchemaSubgoal(schema, "goal_pick_up", SOURCE, PRECONDITION);
    CsInstallOrderingConstraint(schema, "goal_pour", "goal_put_down");
    CsInstallSchemaTriggeringFunctions(schema, milk_from_carton_triggers_object, milk_from_carton_triggers_effector, situation_triggers_milk_from_carton);

    schema = CsInstallSchemaNode(&schemas, "Stir with Stirrer", "stir_it", 2, NONE);
    CsInstallSchemaSubgoal(schema, "goal_put_down", EMPTY, CLEANUP);
    CsInstallSchemaSubgoal(schema, "goal_stir", TARGET, CRUX);
    CsInstallSchemaSubgoal(schema, "goal_pick_up", IMPLEMENT, PRECONDITION);
    CsInstallOrderingConstraint(schema, "goal_stir", "goal_put_down");
    CsInstallSchemaTriggeringFunctions(schema, stir_with_stirrer_triggers_object, stir_with_stirrer_triggers_effector, situation_triggers_stir_with_stirrer);

    schema = CsInstallSchemaNode(&schemas, "Cereal into Mouth", "eat_cereal", 2, NONE);
    CsInstallSchemaSubgoal(schema, "goal_put_down", EMPTY, CLEANUP);
    CsInstallSchemaSubgoal(schema, "goal_consume_spoon_contents", EMPTY, CRUX);
    CsInstallSchemaSubgoal(schema, "goal_dip_spoon", SOURCE, PRECONDITION);
    CsInstallSchemaSubgoal(schema, "goal_pick_up", IMPLEMENT, PRECONDITION);
    CsInstallOrderingConstraint(schema, "goal_consume_spoon_contents", "goal_put_down");
    CsInstallSchemaTriggeringFunctions(schema, cereal_into_mouth_triggers_object, cereal_into_mouth_triggers_effector, situation_triggers_cereal_into_mouth);

    schema = CsInstallSchemaNode(&schemas, "Drink from Mug", "drink_coffee", 2, NONE);
    CsInstallSchemaSubgoal(schema, "goal_put_down", EMPTY, CLEANUP);
    CsInstallSchemaSubgoal(schema, "goal_drink", EMPTY, CRUX);
    CsInstallSchemaSubgoal(schema, "goal_pick_up", SOURCE, PRECONDITION);
    CsInstallOrderingConstraint(schema, "goal_drink", "goal_put_down");
    CsInstallSchemaTriggeringFunctions(schema, drink_from_mug_triggers_object, drink_from_mug_triggers_effector, situation_triggers_drink_from_mug);

/********* Level 1 ************************************************************/

    schema = CsInstallSchemaNode(&schemas, "Pick Up", "goal_pick_up", 1, PICK_UP);
    CsInstallSchemaResource(schema,      HAND);
    CsInstallSchemaTriggeringFunctions(schema, pick_up_triggers_object, pick_up_triggers_effector, situation_triggers_pick_up);
    schema->default_argument_function = ANY;

    schema = CsInstallSchemaNode(&schemas, "Put Down", "goal_put_down", 1, PUT_DOWN);
    CsInstallSchemaResource(schema,      HAND);
    CsInstallSchemaTriggeringFunctions(schema, put_down_triggers_object, put_down_triggers_effector, situation_triggers_put_down);

    schema = CsInstallSchemaNode(&schemas, "Dip Spoon", "goal_dip_spoon", 1, DIP_SPOON);
    CsInstallSchemaResource(schema,      HAND);
    CsInstallSchemaTriggeringFunctions(schema, dip_spoon_triggers_object, dip_spoon_triggers_effector, situation_triggers_dip_spoon);
    schema->default_argument_function = SOURCE;

    schema = CsInstallSchemaNode(&schemas, "Empty Spoon", "goal_empty_spoon", 1, EMPTY_SPOON);
    CsInstallSchemaResource(schema,      HAND);
    CsInstallSchemaTriggeringFunctions(schema, empty_spoon_triggers_object, empty_spoon_triggers_effector, situation_triggers_empty_spoon);
    schema->default_argument_function = TARGET;

    schema = CsInstallSchemaNode(&schemas, "Eat from Spoon", "goal_consume_spoon_contents", 1, EAT_FROM_SPOON);
    CsInstallSchemaResource(schema,      HAND);
    CsInstallSchemaTriggeringFunctions(schema, consume_spoon_contents_triggers_object, consume_spoon_contents_triggers_effector, situation_triggers_consume_spoon_contents);

    schema = CsInstallSchemaNode(&schemas, "Stir", "goal_stir", 1, STIR);
    CsInstallSchemaResource(schema,      HAND);
    CsInstallSchemaTriggeringFunctions(schema, stir_triggers_object, stir_triggers_effector, situation_triggers_stir);
    schema->default_argument_function = ANY;

    schema = CsInstallSchemaNode(&schemas, "Pour", "goal_pour", 1, POUR);
    CsInstallSchemaResource(schema,      HAND);
    CsInstallSchemaTriggeringFunctions(schema, pour_triggers_object, pour_triggers_effector, situation_triggers_pour);
    schema->default_argument_function = TARGET;

    schema = CsInstallSchemaNode(&schemas, "Tear", "goal_open", 1, TEAR);
    CsInstallSchemaResource(schema,      HAND);
    CsInstallSchemaTriggeringFunctions(schema, tear_triggers_object, tear_triggers_effector, situation_triggers_tear);

    schema = CsInstallSchemaNode(&schemas, "Screw Open", "goal_open", 1, SCREW_OPEN);
    CsInstallSchemaResource(schema,      HAND);
    CsInstallSchemaTriggeringFunctions(schema, screw_open_triggers_object, screw_open_triggers_effector, situation_triggers_screw_open);

    schema = CsInstallSchemaNode(&schemas, "Screw Closed", "goal_close", 1, SCREW_CLOSED);
    CsInstallSchemaResource(schema,      HAND);
    CsInstallSchemaTriggeringFunctions(schema, screw_closed_triggers_object, screw_closed_triggers_effector, situation_triggers_screw_closed);

    schema = CsInstallSchemaNode(&schemas, "Swap", "goal_swap", 1, SWAP);
    CsInstallSchemaResource(schema,      HAND);
    CsInstallSchemaTriggeringFunctions(schema, swap_triggers_object, swap_triggers_effector, situation_triggers_swap);

    schema = CsInstallSchemaNode(&schemas, "Drink", "goal_drink", 1, DRINK);
    CsInstallSchemaResource(schema,      HAND);
    CsInstallSchemaTriggeringFunctions(schema, drink_triggers_object, drink_triggers_effector, situation_triggers_drink);


/* Return the pointer to the schema list: */

    return(schemas);
}

/******** END OF FILE *********************************************************/
