/******************************************************************************/

#include "cs.h"

/******************************************************************************/

extern BOOLEAN CoffeeHolds(CsResource *effector, CsObject *object);
extern BOOLEAN CoffeeIsContainer(CsObject *object);
extern BOOLEAN CoffeeContains(CsObject *object, char *functor);
extern BOOLEAN CoffeeIsEmptyContainer(CsObject *source);
extern BOOLEAN CoffeeIsNonemptyContainer(CsObject *source);
extern BOOLEAN CoffeeIsSource(CsObject *object);
extern BOOLEAN CoffeeIsStirrer(CsObject *object);
extern BOOLEAN CoffeeIsSpoon(CsObject *object);
extern BOOLEAN CoffeeIsBowl(CsObject *object);
extern BOOLEAN CoffeeIsJar(CsObject *object);
extern BOOLEAN CoffeeIsMug(CsObject *object);
extern BOOLEAN CoffeeIsPacket(CsObject *object);
extern BOOLEAN CoffeeIsCarton(CsObject *object);

extern BOOLEAN CoffeeIsOpen(CsObject *object);
extern BOOLEAN CoffeeIsClosed(CsObject *object);
extern BOOLEAN CoffeeIsHeld(CsObject *object);
extern BOOLEAN CoffeeIsHand(CsResource *effector);
extern CsResource *CoffeeGetOtherHand(CsResource *hand);
extern BOOLEAN CoffeeIsEmptyHand(CsResource *effector);
extern BOOLEAN CoffeeIsLidOfJar(CsObject *lid, CsObject *jar);
extern BOOLEAN CoffeeIsWithinReach(CsObject *object);
extern BOOLEAN CoffeeHandReaches(CsResource *hand, CsObject *object);
extern BOOLEAN CoffeeNotHeldBySameHand(CsObject *object1, CsObject *object2);

extern TERM *get_feature(GENERIC_LIST *features, char *functor);
extern double max_salience();
extern double net_influence();

/******************************************************************************/

/*-------- Level 4 -----------------------------------------------------------*/

double situation_triggers_breakfast(int cycle)
{
    double total = 0.0; int n = 0;

    total = 1.0; n++;

    return((n==0)?0.0:(total/n));
}

/*-------- Level 3 -----------------------------------------------------------*/

double situation_triggers_coffee(int cycle)
{
    CsObject *mug, *source, *spoon;
    double total = 0.0; int n = 0;

    for (mug = objects; mug != NULL; mug = mug->next) {
        if (CoffeeIsMug(mug)) {
            for (source = objects; source != NULL; source = source->next) {
                if (CoffeeContains(source, "coffee") || CoffeeContains(source, "sugar") || CoffeeContains(source, "milk")) {
                    for (spoon = objects; spoon != NULL; spoon = spoon->next) {
                        if (CoffeeIsSpoon(spoon)) {
                            n++;
                            total += net_influence(3, mug->activation[TARGET][cycle-1], source->activation[SOURCE][cycle-1], spoon->activation[IMPLEMENT][cycle-1]);
                        }
                    }
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_consume_cereal(int cycle)
{
    CsObject *bowl, *source, *spoon;
    double total = 0.0; int n = 0;

    for (bowl = objects; bowl != NULL; bowl = bowl->next) {
        if (CoffeeIsBowl(bowl) && CoffeeContains(bowl, "oatmeal")) {
            for (source = objects; source != NULL; source = source->next) {
                if (CoffeeContains(source, "sugar") || CoffeeContains(source, "milk")) {
                    for (spoon = objects; spoon != NULL; spoon = spoon->next) {
                        if (CoffeeIsSpoon(spoon)) {
                            n++;
                            total += net_influence(4, bowl->activation[SOURCE][cycle-1], bowl->activation[TARGET][cycle-1], source->activation[SOURCE][cycle-1], spoon->activation[IMPLEMENT][cycle-1]);
                        }
                    }
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

/*-------- Level 2 -----------------------------------------------------------*/

double situation_triggers_sugar_from_pack(int cycle)
{
    CsResource *hand, *other_hand;
    CsObject *sugar_pack;
    double total = 0.0; int n = 0;

    for (sugar_pack = objects; sugar_pack != NULL; sugar_pack = sugar_pack->next) {
        if (CoffeeIsPacket(sugar_pack) && CoffeeContains(sugar_pack, "sugar")) {
            for (hand = resources; hand != NULL; hand = hand->next) {
                other_hand = CoffeeGetOtherHand(hand);
                if (CoffeeIsEmptyHand(hand) && CoffeeHandReaches(hand, sugar_pack)) {
                    n++;
                    total += net_influence(3, sugar_pack->activation[SOURCE][cycle-1], hand->activation[cycle-1], other_hand->activation[cycle-1]);
                }
                else if (CoffeeHolds(hand, sugar_pack)) {
                    n++;
                    total += net_influence(3, sugar_pack->activation[SOURCE][cycle-1], hand->activation[cycle-1], other_hand->activation[cycle-1]);
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_sugar_from_bowl(int cycle)
{
    CsObject *bowl, *spoon;
    CsResource *hand;
    double total = 0.0; int n = 0;

    for (bowl = objects; bowl != NULL; bowl = bowl->next) {
        if (CoffeeIsBowl(bowl) && CoffeeContains(bowl, "sugar")) {
            for (spoon = objects; spoon != NULL; spoon = spoon->next) {
                if (CoffeeIsSpoon(spoon) && CoffeeIsEmptyContainer(spoon)) {
                    for (hand = resources; hand != NULL; hand = hand->next) {
                        if (CoffeeIsEmptyHand(hand) && CoffeeHandReaches(hand, spoon) && CoffeeHandReaches(hand, bowl)) {
                            n++;
                            total += net_influence(3, bowl->activation[SOURCE][cycle-1], spoon->activation[IMPLEMENT][cycle-1], hand->activation[cycle-1]);
                        }
                        else if (CoffeeHolds(hand, spoon) && CoffeeHandReaches(hand, bowl)) {
                            n++;
                            total += net_influence(3, bowl->activation[SOURCE][cycle-1], spoon->activation[IMPLEMENT][cycle-1], hand->activation[cycle-1]);
                        }
                    }
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_coffee_from_pack(int cycle)
{
    CsObject *coffee_pack;
    CsResource *hand, *other_hand;
    double total = 0.0; int n = 0;

    for (coffee_pack = objects; coffee_pack != NULL; coffee_pack = coffee_pack->next) {
        if (CoffeeIsPacket(coffee_pack) && CoffeeContains(coffee_pack, "coffee")) {
            for (hand = resources; hand != NULL; hand = hand->next) {
                other_hand = CoffeeGetOtherHand(hand);
                if (CoffeeIsEmptyHand(hand) && CoffeeHandReaches(hand, coffee_pack)) {
                    n++;
                    total += net_influence(3, coffee_pack->activation[SOURCE][cycle-1], hand->activation[cycle-1], other_hand->activation[cycle-1]);
                }
                else if (CoffeeHolds(hand, coffee_pack)) {
                    n++;
                    total += net_influence(3, coffee_pack->activation[SOURCE][cycle-1], hand->activation[cycle-1], other_hand->activation[cycle-1]);
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_coffee_from_jar(int cycle)
{
    CsObject *jar, *spoon;
    CsResource *hand;
    double total = 0.0; int n = 0;

    for (jar = objects; jar != NULL; jar = jar->next) {
        if (CoffeeIsJar(jar) && CoffeeContains(jar, "coffee")) {
            for (spoon = objects; spoon != NULL; spoon = spoon->next) {
                if (CoffeeIsSpoon(spoon) && CoffeeIsEmptyContainer(spoon)) {
                    for (hand = resources; hand != NULL; hand = hand->next) {
                        if (CoffeeIsEmptyHand(hand) && CoffeeHandReaches(hand, spoon) && CoffeeHandReaches(hand, jar)) {
                            n++;
                            total += net_influence(3, jar->activation[SOURCE][cycle-1], spoon->activation[IMPLEMENT][cycle-1], hand->activation[cycle-1]);
                        }
                        else if (CoffeeHolds(hand, spoon) && CoffeeHandReaches(hand, jar)) {
                            n++;
                            total += net_influence(3, jar->activation[SOURCE][cycle-1], spoon->activation[IMPLEMENT][cycle-1], hand->activation[cycle-1]);
                        }
                    }
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_milk_from_carton(int cycle)
{
    CsObject *carton;
    CsResource *hand, *other_hand;
    double total = 0.0; int n = 0;

    for (carton = objects; carton != NULL; carton = carton->next) {
        if (CoffeeIsCarton(carton) && CoffeeContains(carton, "milk")) {
            for (hand = resources; hand != NULL; hand = hand->next) {
                other_hand = CoffeeGetOtherHand(hand);
                if (CoffeeIsEmptyHand(hand) && CoffeeHandReaches(hand, carton)) {
                    n++;
                    total += net_influence(3, carton->activation[SOURCE][cycle-1], hand->activation[cycle-1], other_hand->activation[cycle-1]);
                }
                else if (CoffeeHolds(hand, carton)) {
                    n++;
                    total += net_influence(3, carton->activation[SOURCE][cycle-1], hand->activation[cycle-1], other_hand->activation[cycle-1]);
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_stir_with_stirrer(int cycle)
{
    CsObject *source, *stirrer;
    CsResource *hand;
    double total = 0.0; int n = 0;

    for (source = objects; source != NULL; source = source->next) {
        if (CoffeeIsContainer(source) && CoffeeIsOpen(source)) {
            for (stirrer = objects; stirrer != NULL; stirrer = stirrer->next) {
                if (CoffeeIsStirrer(stirrer)) {
                    for (hand = resources; hand != NULL; hand = hand->next) {
                        if (CoffeeIsEmptyHand(hand) && CoffeeHandReaches(hand, stirrer) && CoffeeHandReaches(hand, source)) {
                            n++;
                            total += net_influence(3, source->activation[TARGET][cycle-1], stirrer->activation[IMPLEMENT][cycle-1], hand->activation[cycle-1]);
                        }
                        else if (CoffeeHolds(hand, stirrer) && CoffeeHandReaches(hand, source)) {
                            n++;
                            total += net_influence(3, source->activation[TARGET][cycle-1], stirrer->activation[IMPLEMENT][cycle-1], hand->activation[cycle-1]);
                        }
                    }
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_cereal_into_mouth(int cycle)
{
    CsObject *bowl, *spoon;
    CsResource *hand;
    double total = 0.0; int n = 0;

    for (bowl = objects; bowl != NULL; bowl = bowl->next) {
        if (CoffeeIsBowl(bowl) && CoffeeContains(bowl, "oatmeal")) {
            for (spoon = objects; spoon != NULL; spoon = spoon->next) {
                if (CoffeeIsSpoon(spoon) && CoffeeIsEmptyContainer(spoon)) {
                    for (hand = resources; hand != NULL; hand = hand->next) {
                        if (CoffeeIsEmptyHand(hand) && CoffeeHandReaches(hand, spoon) && CoffeeHandReaches(hand, bowl)) {
                            n++;
                            total += net_influence(3, bowl->activation[SOURCE][cycle-1], spoon->activation[IMPLEMENT][cycle-1], hand->activation[cycle-1]);
                        }
                        else if (CoffeeHolds(hand, spoon) && CoffeeHandReaches(hand, bowl)) {
                            n++;
                            total += net_influence(3, bowl->activation[SOURCE][cycle-1], spoon->activation[IMPLEMENT][cycle-1], hand->activation[cycle-1]);
                        }
                    }
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_drink_from_mug(int cycle)
{
    CsObject *mug;
    CsResource *hand;
    double total = 0.0; int n = 0;

    for (mug = objects; mug != NULL; mug = mug->next) {
        if (CoffeeIsMug(mug) && CoffeeIsNonemptyContainer(mug)) {
            for (hand = resources; hand != NULL; hand = hand->next) {
                if (CoffeeIsEmptyHand(hand) && CoffeeHandReaches(hand, mug)) {
                    n++;
                    total += net_influence(2, mug->activation[SOURCE][cycle-1], hand->activation[cycle-1]);
                }
                else if (CoffeeHolds(hand, mug)) {
                    n++;
                    total += net_influence(2, mug->activation[SOURCE][cycle-1], hand->activation[cycle-1]);
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

/*-------- Level 1 -----------------------------------------------------------*/

double situation_triggers_pick_up(int cycle)
{
    CsObject *object;
    CsResource *hand;
    double total = 0.0; int n = 0;

    for (hand = resources; hand != NULL; hand = hand->next) {
        if (CoffeeIsEmptyHand(hand)) {
            for (object = objects; object != NULL; object = object->next) {
                if (CoffeeHandReaches(hand, object)) {
                    n++;
                    total += net_influence(2, object->activation[IMPLEMENT][cycle-1], hand->activation[cycle-1]) +
                             net_influence(2, object->activation[SOURCE][cycle-1], hand->activation[cycle-1]) +
                             net_influence(2, object->activation[TARGET][cycle-1], hand->activation[cycle-1]);
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_put_down(int cycle)
{
    CsObject *object;
    CsResource *hand;
    double total = 0.0; int n = 0;

    for (object = objects; object != NULL; object = object->next) {
        for (hand = resources; hand != NULL; hand = hand->next) {
            if (CoffeeIsHand(hand) && CoffeeHolds(hand, object)) {
                n++;
                total += net_influence(2, object->activation[IMPLEMENT][cycle-1], hand->activation[cycle-1]) +
                         net_influence(2, object->activation[SOURCE][cycle-1], hand->activation[cycle-1]) +
                         net_influence(2, object->activation[TARGET][cycle-1], hand->activation[cycle-1]);
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_dip_spoon(int cycle)
{
    CsObject *spoon, *source;
    CsResource *hand;
    double total = 0.0; int n = 0;

    for (spoon = objects; spoon != NULL; spoon = spoon->next) {
        if (CoffeeIsSpoon(spoon) && CoffeeIsEmptyContainer(spoon)) {
            for (source = objects; source != NULL; source = source->next) {
                if (CoffeeIsOpen(source) && CoffeeIsSource(source)) {
                    for (hand = resources; hand != NULL; hand = hand->next) {
                        if (CoffeeHolds(hand, spoon) && CoffeeHandReaches(hand, source)) {
                            n++;
                            total += net_influence(3, source->activation[SOURCE][cycle-1], spoon->activation[IMPLEMENT][cycle-1], hand->activation[cycle-1]);
                        }
                    }
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_empty_spoon(int cycle)
{
    CsObject *target, *spoon;
    CsResource *hand;
    double total = 0.0; int n = 0;

    for (spoon = objects; spoon != NULL; spoon = spoon->next) {
        if (CoffeeIsSpoon(spoon) && CoffeeIsNonemptyContainer(spoon)) {
            for (target = objects; target != NULL; target = target->next) {
                if (CoffeeIsOpen(target) && CoffeeIsContainer(target)) {
                    for (hand = resources; hand != NULL; hand = hand->next) {
                        if (CoffeeHolds(hand, spoon) && CoffeeHandReaches(hand, target)) {
                            n++;
                            total += net_influence(3, target->activation[TARGET][cycle-1], spoon->activation[IMPLEMENT][cycle-1], hand->activation[cycle-1]);
                        }
                    }
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_consume_spoon_contents(int cycle)
{
    CsObject *spoon;
    CsResource *hand;
    double total = 0.0; int n = 0;

    for (spoon = objects; spoon != NULL; spoon = spoon->next) {
        if (CoffeeIsSpoon(spoon) && CoffeeIsNonemptyContainer(spoon)) {
            for (hand = resources; hand != NULL; hand = hand->next) {
                if (CoffeeHolds(hand, spoon)) {
                    n++;
                    total += net_influence(2, spoon->activation[IMPLEMENT][cycle-1], hand->activation[cycle-1]);
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_stir(int cycle)
{
    CsObject *stirrer, *source;
    CsResource *hand;
    double total = 0.0; int n = 0;

    for (stirrer = objects; stirrer != NULL; stirrer = stirrer->next) {
        if (CoffeeIsStirrer(stirrer)) {
            for (source = objects; source != NULL; source = source->next) {
                if (CoffeeIsOpen(source) && CoffeeIsContainer(source) && CoffeeIsNonemptyContainer(source)) {
                    for (hand = resources; hand != NULL; hand = hand->next) {
                        if (CoffeeHolds(hand, stirrer) && CoffeeHandReaches(hand, source)) {
                            n++;
                            total += net_influence(3, source->activation[SOURCE][cycle-1], stirrer->activation[IMPLEMENT][cycle-1], hand->activation[cycle-1]);
                        }
                    }
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_pour(int cycle)
{
    CsObject *source, *target;
    CsResource *hand;
    double total = 0.0; int n = 0;

    for (source = objects; source != NULL; source = source->next) {
        if (CoffeeIsOpen(source) && CoffeeIsSource(source)) {
            for (target = objects; target != NULL; target = target->next) {
                if (CoffeeIsOpen(target) && CoffeeIsContainer(target) && CoffeeNotHeldBySameHand(source, target)) {
                    for (hand = resources; hand != NULL; hand = hand ->next) {
                        if (CoffeeHolds(hand, source) && CoffeeHandReaches(hand, target)) {
                            n++;
                            total += net_influence(3, source->activation[SOURCE][cycle-1], target->activation[TARGET][cycle-1], hand->activation[cycle-1]);
                        }
                    }
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_tear(int cycle)
{
    CsObject *source;
    CsResource *hand;
    double total = 0.0; int n = 0;

    for (source = objects; source != NULL; source = source->next) {
        if (CoffeeIsHeld(source) && (CoffeeIsPacket(source) || CoffeeIsCarton(source)) && CoffeeIsClosed(source)) {
            for (hand = resources; hand != NULL; hand = hand->next) {
                if (CoffeeIsEmptyHand(hand)) {
                    n++;
                    total += net_influence(2, source->activation[SOURCE][cycle-1], hand->activation[cycle-1]);
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_screw_open(int cycle)
{
    CsObject *jar;
    CsResource *hand1, *hand2;
    double total = 0.0; int n = 0;

    for (jar = objects; jar != NULL; jar = jar->next) {
        if (CoffeeIsJar(jar) && CoffeeIsClosed(jar)) {
            for (hand1 = resources; hand1 != NULL; hand1 = hand1->next) {
                if (CoffeeIsHand(hand1) && CoffeeHolds(hand1, jar)) {
                    for (hand2 = resources; hand2 != NULL; hand2 = hand2->next) {
                        if (CoffeeIsEmptyHand(hand2)) {
                            n++;
                            total += net_influence(3, jar->activation[SOURCE][cycle-1], hand1->activation[cycle-1], hand2->activation[cycle-1]);
                        }
                    }
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_screw_closed(int cycle)
{
    CsObject *jar, *lid;
    CsResource *hand1, *hand2;
    double total = 0.0; int n = 0;

    for (jar = objects; jar != NULL; jar = jar->next) {
        if (CoffeeIsJar(jar) && CoffeeIsOpen(jar)) {
            for (lid = objects; lid != NULL; lid = lid->next) {
                if (CoffeeIsLidOfJar(lid, jar)) {
                    for (hand1 = resources; hand1 != NULL; hand1 = hand1->next) {
                        if (CoffeeIsHand(hand1) && CoffeeHolds(hand1, jar)) {
                            for (hand2 = resources; hand2 != NULL; hand2 = hand2->next) {
                                if (CoffeeIsHand(hand2) && CoffeeHolds(hand2, lid)) {
                                    n++;
                                    total += net_influence(4, max_salience(lid, cycle-1), jar->activation[SOURCE][cycle-1], hand1->activation[cycle-1], hand2->activation[cycle-1]);
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_swap(int cycle)
{
    CsObject *object;
    CsResource *full_hand, *empty_hand;
    double total = 0.0; int n = 0;

    for (full_hand = resources; full_hand != NULL; full_hand = full_hand->next) {
        if (CoffeeIsHand(full_hand)) {
            for (object = objects; object != NULL; object = object->next) {
                if (CoffeeHolds(full_hand, object)) {
                    for (empty_hand = resources; empty_hand != NULL; empty_hand = empty_hand->next) {
                        if (CoffeeIsEmptyHand(empty_hand)) {
                            n++;
                            total += net_influence(3, max_salience(object, cycle-1), empty_hand->activation[cycle-1], full_hand->activation[cycle-1]);
                        }
                    }
                }
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

double situation_triggers_drink(int cycle)
{
    CsObject *object;
    CsResource *hand;
    double total = 0.0; int n = 0;

    for (object = objects; object != NULL; object = object->next) {
        for (hand = resources; hand != NULL; hand = hand->next) {
            if (CoffeeIsHand(hand) && CoffeeHolds(hand, object) && CoffeeIsNonemptyContainer(object)) {
                n++;
                total += net_influence(2, object->activation[SOURCE][cycle-1], hand->activation[cycle-1]);
            }
        }
    }
    return((n==0)?0.0:(total/n));
}

/******************************************************************************/

/*-------- Level 4 -----------------------------------------------------------*/

BOOLEAN breakfast_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    return(TRUE);
}

/*-------- Level 3 -----------------------------------------------------------*/

BOOLEAN coffee_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsObject *source, *spoon, *mug;

    if ((function == TARGET) && CoffeeIsMug(object)) {
        for (source = objects; source != NULL; source = source->next) {
            if (CoffeeContains(source, "coffee") || CoffeeContains(source, "sugar") || CoffeeContains(source, "milk")) {
                for (spoon = objects; spoon != NULL; spoon = spoon->next) {
                    if (CoffeeIsSpoon(spoon)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    else if ((function == SOURCE) && (CoffeeContains(object, "coffee") || CoffeeContains(object, "sugar") || CoffeeContains(object, "milk"))){
        for (mug = objects; mug != NULL; mug = mug->next) {
            if (CoffeeIsMug(mug)) {
                for (spoon = objects; spoon != NULL; spoon = spoon->next) {
                    if (CoffeeIsSpoon(spoon)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    else if ((function == IMPLEMENT) && CoffeeIsSpoon(object)) {
        for (mug = objects; mug != NULL; mug = mug->next) {
            if (CoffeeIsMug(mug)) {
                for (source = objects; (source != NULL); source = source->next) {
                    if (CoffeeContains(source, "coffee") || CoffeeContains(source, "sugar") || CoffeeContains(source, "milk")) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN consume_cereal_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsObject *spoon, *source, *bowl;

    if (((function == SOURCE) || (function == TARGET)) && CoffeeIsBowl(object) && CoffeeContains(object, "oatmeal")) {
        for (source = objects; source != NULL; source = source->next) {
            if (CoffeeContains(source, "sugar") || CoffeeContains(source, "milk")) {
                for (spoon = objects; spoon != NULL; spoon = spoon->next) {
                    if (CoffeeIsSpoon(spoon)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    else if ((function == SOURCE) && (CoffeeContains(object, "sugar") || CoffeeContains(object, "milk"))){
        for (bowl = objects; bowl != NULL; bowl = bowl->next) {
            if (CoffeeIsBowl(bowl) && CoffeeContains(bowl, "oatmeal")) {
                for (spoon = objects; spoon != NULL; spoon = spoon->next) {
                    if (CoffeeIsSpoon(spoon)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    else if ((function == IMPLEMENT) && CoffeeIsSpoon(object)) {
        for (bowl = objects; bowl != NULL; bowl = bowl->next) {
            if (CoffeeIsBowl(bowl) && CoffeeContains(bowl, "oatmeal")) {
                for (source = objects; (source != NULL); source = source->next) {
                    if (CoffeeContains(source, "coffee") || CoffeeContains(source, "sugar") || CoffeeContains(source, "milk")) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    return(FALSE);
}

/*-------- Level 2 -----------------------------------------------------------*/

BOOLEAN sugar_from_pack_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    if ((function == SOURCE) && CoffeeIsPacket(object) && CoffeeContains(object, "sugar")) {
        if (CoffeeIsHeld(object) || CoffeeIsWithinReach(object)) {
            return(TRUE);
        }
    }
    return(FALSE);
}

BOOLEAN sugar_from_bowl_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsObject *spoon, *bowl;
    CsResource *hand;

    if ((function == SOURCE) && CoffeeIsBowl(object) && CoffeeContains(object, "sugar")) {
        for (spoon = objects; spoon != NULL; spoon = spoon->next) {
            if (CoffeeIsSpoon(spoon) && CoffeeIsEmptyContainer(spoon)) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if ((CoffeeHolds(hand, spoon) || CoffeeHandReaches(hand, spoon)) && CoffeeHandReaches(hand, object)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    else if ((function == IMPLEMENT) && CoffeeIsSpoon(object) && CoffeeIsEmptyContainer(object)) {
        for (bowl = objects; bowl != NULL; bowl = bowl->next) {
            if (CoffeeIsBowl(bowl) && CoffeeContains(bowl, "sugar")) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if ((CoffeeHolds(hand, object) || CoffeeHandReaches(hand, object)) && CoffeeHandReaches(hand, bowl)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN coffee_from_pack_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    if ((function == SOURCE) && CoffeeIsPacket(object) && CoffeeContains(object, "coffee")) {
        if (CoffeeIsHeld(object) || CoffeeIsWithinReach(object)) {
            return(TRUE);
        }
    }
    return(FALSE);
}

BOOLEAN coffee_from_jar_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsObject *spoon, *jar;
    CsResource *hand;

    if ((function == SOURCE) && CoffeeIsJar(object) && CoffeeContains(object, "coffee")) {
        for (spoon = objects; spoon != NULL; spoon = spoon->next) {
            if (CoffeeIsSpoon(spoon) && CoffeeIsEmptyContainer(spoon)) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if ((CoffeeHolds(hand, spoon) || CoffeeHandReaches(hand, spoon)) && CoffeeHandReaches(hand, spoon)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    else if ((function == IMPLEMENT) && CoffeeIsSpoon(object) && CoffeeIsEmptyContainer(object)) {
        for (jar = objects; jar != NULL; jar = jar->next) {
            if (CoffeeIsJar(jar) && CoffeeContains(jar, "coffee")) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if ((CoffeeHolds(hand, object) || CoffeeHandReaches(hand, object)) && CoffeeHandReaches(hand, jar)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN milk_from_carton_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    if ((function == SOURCE) && CoffeeIsCarton(object) && CoffeeContains(object, "milk")) {
        if (CoffeeIsHeld(object) || CoffeeIsWithinReach(object)) {
            return(TRUE);
        }
    }
    return(FALSE);
}

BOOLEAN stir_with_stirrer_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsObject *target, *stirrer;
    CsResource *hand;

    if ((function == IMPLEMENT) && CoffeeIsStirrer(object)) {
        for (target = objects; target != NULL; target = target->next) {
            if (CoffeeIsOpen(target) && CoffeeIsContainer(target)) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if ((CoffeeHolds(hand, object) || CoffeeHandReaches(hand, object)) && CoffeeHandReaches(hand, target)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    else if ((function == TARGET) && CoffeeIsContainer(object) && CoffeeIsOpen(object)) {
        for (stirrer = objects; stirrer != NULL; stirrer = stirrer->next) {
            if (CoffeeIsStirrer(stirrer)) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if ((CoffeeHolds(hand, stirrer) || CoffeeHandReaches(hand, stirrer)) && CoffeeHandReaches(hand, object)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN cereal_into_mouth_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsObject *spoon, *bowl;
    CsResource *hand;

    if ((function == SOURCE) && CoffeeIsBowl(object) && CoffeeContains(object, "oatmeal")) {
        for (spoon = objects; spoon != NULL; spoon = spoon->next) {
            if (CoffeeIsSpoon(spoon) && (CoffeeIsEmptyContainer(spoon) || CoffeeContains(spoon, "oatmeal"))) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if ((CoffeeHolds(hand, spoon) || CoffeeHandReaches(hand, spoon)) && CoffeeHandReaches(hand, object)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    else if ((function == IMPLEMENT) && CoffeeIsSpoon(object) && (CoffeeIsEmptyContainer(object) || CoffeeContains(object, "oatmeal"))) {
        for (bowl = objects; bowl != NULL; bowl = bowl->next) {
            if (CoffeeIsBowl(bowl) && CoffeeContains(bowl, "oatmeal")) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if ((CoffeeHolds(hand, object) || CoffeeHandReaches(hand, object)) && CoffeeHandReaches(hand, bowl)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN drink_from_mug_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    if ((function == SOURCE) && CoffeeIsMug(object) && CoffeeIsNonemptyContainer(object)) {
        if (CoffeeIsHeld(object) || CoffeeIsWithinReach(object)) {
            return(TRUE);
        }
    }
    return(FALSE);
}

/*-------- Level 1 -----------------------------------------------------------*/

BOOLEAN pick_up_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsResource *hand;

    for (hand = resources; hand != NULL; hand = hand->next) {
        if (CoffeeIsEmptyHand(hand) && CoffeeHandReaches(hand, object)) {
            return(TRUE);
        }
    }
    return(FALSE);
}    

BOOLEAN put_down_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    if (CoffeeIsHeld(object)) {
        return(TRUE);
    }
    return(FALSE);
}

BOOLEAN dip_spoon_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsObject *spoon, *source;
    CsResource *hand;

    if ((function == SOURCE) && CoffeeIsOpen(object) && CoffeeIsSource(object)) {
        for (spoon = objects; spoon != NULL; spoon = spoon->next) {
            if (CoffeeIsSpoon(spoon) && CoffeeIsEmptyContainer(spoon)) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if (CoffeeHolds(hand, spoon) && CoffeeHandReaches(hand, object)) {
                        return(TRUE);
                    }
                }
            }
        }
        return(FALSE);
    }
    else if ((function == IMPLEMENT) && CoffeeIsSpoon(object) && CoffeeIsEmptyContainer(object)) {
        for (source = objects; source != NULL; source = source->next) {
            if (CoffeeIsOpen(source) && CoffeeIsSource(source)) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if (CoffeeHolds(hand, object) && CoffeeHandReaches(hand, source)) {
                        return(TRUE);
                    }
                }
            }
        }
        return(FALSE);
    }
    else {
        return(FALSE);
    }
}

BOOLEAN empty_spoon_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsObject *spoon, *target;
    CsResource *hand;

    if ((function == TARGET) && CoffeeIsOpen(object) && CoffeeIsContainer(object)) {
        for (spoon = objects; spoon != NULL; spoon = spoon->next) {
            if (CoffeeIsSpoon(spoon) && CoffeeIsNonemptyContainer(spoon)) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if (CoffeeHolds(hand, spoon) && CoffeeHandReaches(hand, object)) {
                        return(TRUE);
                    }
                }
            }
        }
        return(FALSE);
    }
    else if ((function == IMPLEMENT) && CoffeeIsSpoon(object) && CoffeeIsNonemptyContainer(object)) {
        for (target = objects; target != NULL; target = target->next) {
            if (CoffeeIsOpen(target) && CoffeeIsContainer(target)) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if (CoffeeHolds(hand, object) && CoffeeHandReaches(hand, target)) {
                        return(TRUE);
                    }
                }
            }
        }
        return(FALSE);
    }
    else {
        return(FALSE);
    }
}

BOOLEAN consume_spoon_contents_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsResource *hand;

    if ((function == IMPLEMENT) && CoffeeIsSpoon(object) && CoffeeIsNonemptyContainer(object)) {
        for (hand = resources; hand != NULL; hand = hand->next) {
            if (CoffeeHolds(hand, object)) {
                return(TRUE);
            }
        }
        return(FALSE);
    }
    else {
        return(FALSE);
    }
}

BOOLEAN stir_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsObject *stirrer, *source;
    CsResource *hand;

    if ((function == SOURCE) && CoffeeIsOpen(object) && CoffeeIsContainer(object) && CoffeeIsNonemptyContainer(object)) {
        for (stirrer = objects; stirrer != NULL; stirrer = stirrer->next) {
            if (CoffeeIsStirrer(stirrer)) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if (CoffeeHolds(hand, stirrer) && CoffeeHandReaches(hand, object)) {
                        return(TRUE);
                    }
                }
            }
        }
        return(FALSE);
    }
    else if ((function == IMPLEMENT) && CoffeeIsStirrer(object)) {
        for (source = objects; source != NULL; source = source->next) {
            if (CoffeeIsOpen(source) && CoffeeIsContainer(source) && CoffeeIsNonemptyContainer(source)) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if (CoffeeHolds(hand, object) && CoffeeHandReaches(hand, source)) {
                        return(TRUE);
                    }
                }
            }
        }
        return(FALSE);
    }
    else {
        return(FALSE);
    }
}

BOOLEAN pour_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsObject *target, *source;
    CsResource *hand;

    if ((function == SOURCE) && CoffeeIsOpen(object) && CoffeeIsSource(object)) {
        for (target = objects; target != NULL; target = target->next) {
            if (CoffeeIsOpen(target) && CoffeeIsContainer(target) && CoffeeNotHeldBySameHand(object, target)) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if (CoffeeHolds(hand, object) && CoffeeHandReaches(hand, target)) {
                        return(TRUE);
                    }
                }
            }
        }
        return(FALSE);
    }
    else if ((function == TARGET) && CoffeeIsOpen(object) && CoffeeIsContainer(object)) {
        for (source = objects; source != NULL; source = source->next) {
            if (CoffeeIsOpen(source) && CoffeeIsSource(source) && CoffeeNotHeldBySameHand(source, object)) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if (CoffeeHolds(hand, source) && CoffeeHandReaches(hand, object)) {
                        return(TRUE);
                    }
                }
            }
        }
        return(FALSE);
    }
    else {
        return(FALSE);
    }
}

BOOLEAN tear_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsResource *hand;

    if ((function == SOURCE) && CoffeeIsHeld(object) && (CoffeeIsPacket(object) || CoffeeIsCarton(object)) && CoffeeIsClosed(object)) {
        for (hand = resources; hand != NULL; hand = hand->next) {
            if (CoffeeIsEmptyHand(hand)) {
                return(TRUE);
            }
        }
    }
    return(FALSE);
}

BOOLEAN screw_open_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsResource *empty_hand, *full_hand;

    if ((function == SOURCE) && CoffeeIsJar(object) && CoffeeIsClosed(object)) {
        for (empty_hand = resources; empty_hand != NULL; empty_hand = empty_hand->next) {
            if (CoffeeIsEmptyHand(empty_hand)) {
                for (full_hand = resources; full_hand != NULL; full_hand = full_hand->next) {
                    if (CoffeeIsHand(full_hand) && CoffeeHolds(full_hand, object)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN screw_closed_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsObject *lid, *jar;
    CsResource *jar_hand, *lid_hand;

    if ((function == SOURCE) && CoffeeIsJar(object) && CoffeeIsOpen(object)) {
        for (lid = objects; lid != NULL; lid = lid->next) {
            if (CoffeeIsLidOfJar(lid, object)) {
                for (jar_hand = resources; jar_hand != NULL; jar_hand = jar_hand->next) {
                    if (CoffeeIsHand(jar_hand) && CoffeeHolds(jar_hand, object)) {
                        for (lid_hand = resources; lid_hand != NULL; lid_hand = lid_hand->next) {
                            if (CoffeeIsHand(lid_hand) && CoffeeHolds(lid_hand, lid)) {
                                return(TRUE);
                            }
                        }
                    }
                }
            }
        }
    }
    else {
        for (jar = objects; jar != NULL; jar = jar->next) {
            if (CoffeeIsJar(jar) && CoffeeIsLidOfJar(object, jar) && CoffeeIsOpen(jar)) {
                for (jar_hand = resources; jar_hand != NULL; jar_hand = jar_hand->next) {
                    if (CoffeeIsHand(jar_hand) && CoffeeHolds(jar_hand, jar)) {
                        for (lid_hand = resources; lid_hand != NULL; lid_hand = lid_hand->next) {
                            if (CoffeeIsHand(lid_hand) && CoffeeHolds(lid_hand, object)) {
                                return(TRUE);
                            }
                        }
                    }
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN swap_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    CsResource *empty_hand, *full_hand;

    for (full_hand = resources; full_hand != NULL; full_hand = full_hand->next) {
        if (CoffeeIsHand(full_hand) && CoffeeHolds(full_hand, object)) {
            for (empty_hand = resources; empty_hand != NULL; empty_hand = empty_hand->next) {
                if (CoffeeIsEmptyHand(empty_hand)) {
                    return(TRUE);
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN drink_triggers_object(CsObject *object, ArgumentFunction function, int cycle)
{
    if (CoffeeIsHeld(object) && CoffeeIsNonemptyContainer(object)) {
        return(TRUE);
    }
    return(FALSE);
}

/******************************************************************************/

/*-------- Level 4 -----------------------------------------------------------*/

BOOLEAN breakfast_triggers_effector(CsResource *effector, int cycle)
{
    return(FALSE);

}

/*-------- Level 3 -----------------------------------------------------------*/

BOOLEAN coffee_triggers_effector(CsResource *effector, int cycle)
{
    return(FALSE);

}

BOOLEAN consume_cereal_triggers_effector(CsResource *effector, int cycle)
{
    return(FALSE);
}

/*-------- Level 2 -----------------------------------------------------------*/

BOOLEAN sugar_from_pack_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *pack;

    for (pack = objects; pack != NULL; pack = pack->next) {
        if (CoffeeIsPacket(pack) && CoffeeContains(pack, "sugar")) {
            if (CoffeeIsEmptyHand(effector) && CoffeeHandReaches(effector, pack)) {
                return(TRUE);
            }
            else if (CoffeeIsHeld(pack)) {
                return(TRUE);
            }
        }
    }
    return(FALSE);
}

BOOLEAN sugar_from_bowl_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *spoon, *bowl;

    for (spoon = objects; spoon != NULL; spoon = spoon->next) {
        if (CoffeeIsSpoon(spoon)) {
            for (bowl = objects; bowl != NULL; bowl = bowl->next) {
                if (CoffeeIsBowl(bowl) && CoffeeHandReaches(effector, bowl) && CoffeeContains(bowl, "sugar")) {
                    if (CoffeeIsEmptyHand(effector) && CoffeeHandReaches(effector, spoon)) {
                        return(TRUE);
                    }
                    else if (CoffeeHolds(effector, spoon)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN coffee_from_pack_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *pack;

    for (pack = objects; pack != NULL; pack = pack->next) {
        if (CoffeeIsPacket(pack) && CoffeeContains(pack, "coffee")) {
            if (CoffeeIsEmptyHand(effector) && CoffeeHandReaches(effector, pack)) {
                return(TRUE);
            }
            else if (CoffeeIsHeld(pack)) {
                return(TRUE);
            }
	}
    }
    return(FALSE);
}

BOOLEAN coffee_from_jar_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *spoon, *jar;

    for (spoon = objects; spoon != NULL; spoon = spoon->next) {
        if (CoffeeIsSpoon(spoon)) {
            for (jar = objects; jar != NULL; jar = jar->next) {
                if (CoffeeIsJar(jar) && CoffeeHandReaches(effector, jar) && CoffeeContains(jar, "coffee")) {
                    if (CoffeeIsEmptyHand(effector) && CoffeeHandReaches(effector, spoon)) {
                        return(TRUE);
                    }
                    else if (CoffeeHolds(effector, spoon)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN milk_from_carton_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *carton;

    for (carton = objects; carton != NULL; carton = carton->next) {
        if (CoffeeIsCarton(carton) && CoffeeContains(carton, "milk")) {
            if (CoffeeIsEmptyHand(effector) && CoffeeHandReaches(effector, carton)) {
                return(TRUE);
            }
            else if (CoffeeIsHeld(carton)) {
                return(TRUE);
            }
	}
    }
    return(FALSE);
}

BOOLEAN stir_with_stirrer_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *stirrer, *target;

    for (stirrer = objects; stirrer != NULL; stirrer = stirrer->next) {
        if (CoffeeIsSpoon(stirrer)) {
            for (target = objects; target != NULL; target = target->next) {
                if (CoffeeIsContainer(target) && CoffeeIsOpen(target) && CoffeeHandReaches(effector, target)) {
                    if (CoffeeIsEmptyHand(effector) && CoffeeHandReaches(effector, stirrer)) {
                        return(TRUE);
                    }
                    else if (CoffeeHolds(effector, stirrer)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN cereal_into_mouth_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *bowl, *spoon;

    for (spoon = objects; spoon != NULL; spoon = spoon->next) {
        if (CoffeeIsSpoon(spoon)) {
            for (bowl = objects; bowl != NULL; bowl = bowl->next) {
                if (CoffeeIsContainer(bowl) && CoffeeIsOpen(bowl) && CoffeeContains(bowl, "oatmeal") && CoffeeHandReaches(effector, bowl)) {
                    if (CoffeeIsEmptyHand(effector) && CoffeeHandReaches(effector, spoon)) {
                        return(TRUE);
                    }
                    else if (CoffeeHolds(effector, spoon)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN drink_from_mug_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *mug;

    for (mug = objects; mug != NULL; mug = mug->next) {
        if (CoffeeIsMug(mug) && CoffeeIsNonemptyContainer(mug)) {
            if (CoffeeIsEmptyHand(effector) && CoffeeHandReaches(effector, mug)) {
                return(TRUE);
            }
            else if (CoffeeIsHeld(mug)) {
                return(TRUE);
            }
	}
    }
    return(FALSE);
}

/*-------- Level 1 -----------------------------------------------------------*/

BOOLEAN pick_up_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *object;

    if (CoffeeIsEmptyHand(effector)) {
        for (object = objects; object != NULL; object = object->next) {
            if (CoffeeHandReaches(effector, object)) {
                return(TRUE);
            }
        }
    }
    return(FALSE);
}

BOOLEAN put_down_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *object;

    if (CoffeeIsHand(effector)) {
        for (object = objects; object != NULL; object = object->next) {
            if (CoffeeHolds(effector, object)) {
                return(TRUE);
            }
        }
    }
    return(FALSE);
}

BOOLEAN dip_spoon_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *spoon, *source;

    for (spoon = objects; spoon != NULL; spoon = spoon->next) {
        if (CoffeeIsSpoon(spoon) && CoffeeIsEmptyContainer(spoon) && CoffeeHolds(effector, spoon)) {
            for (source = objects; source != NULL; source = source->next) {
                if (CoffeeIsOpen(source) && CoffeeIsSource(source) && CoffeeHandReaches(effector, source)) {
                    return(TRUE);
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN empty_spoon_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *spoon, *target;

    for (spoon = objects; spoon != NULL; spoon = spoon->next) {
        if (CoffeeHolds(effector, spoon) && CoffeeIsSpoon(spoon) && CoffeeIsNonemptyContainer(spoon)) {
            for (target = objects; target != NULL; target = target->next) {
                if (CoffeeIsOpen(target) && CoffeeIsContainer(target) && CoffeeHandReaches(effector, target)) {
                    return(TRUE);
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN consume_spoon_contents_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *spoon;

    for (spoon = objects; spoon != NULL; spoon = spoon->next) {
        if (CoffeeHolds(effector, spoon) && CoffeeIsSpoon(spoon) && CoffeeIsNonemptyContainer(spoon)) {
            return(TRUE);
        }
    }
    return(FALSE);
}

BOOLEAN stir_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *stirrer, *source;

    for (stirrer = objects; stirrer != NULL; stirrer = stirrer->next) {
        if (CoffeeIsStirrer(stirrer) && CoffeeHolds(effector, stirrer)) {
            for (source = objects; source != NULL; source = source->next) {
                if (CoffeeIsOpen(source) && CoffeeIsContainer(source) && CoffeeIsNonemptyContainer(source) && CoffeeHandReaches(effector, source)) {
                    return(TRUE);
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN pour_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *target, *source;

    for (target = objects; target != NULL; target = target->next) {
        if (CoffeeIsOpen(target) && CoffeeIsContainer(target) && CoffeeHandReaches(effector, target)) {
            for (source = objects; source != NULL; source = source->next) {
                if (CoffeeHolds(effector, source) && CoffeeIsOpen(source) && CoffeeIsSource(source) && CoffeeNotHeldBySameHand(source, target)) {
                    return(TRUE);
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN tear_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *source;

    if (CoffeeIsEmptyHand(effector)) {
        for (source = objects; source != NULL; source = source->next) {
            if (CoffeeIsHeld(source) && (CoffeeIsPacket(source) || CoffeeIsCarton(source)) && CoffeeIsClosed(source)) {
                return(TRUE);
            }
        }
    }
    return(FALSE);
}

BOOLEAN screw_open_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *jar;
    CsResource *hand;

    if (CoffeeIsEmptyHand(effector)) {
        for (jar = objects; jar != NULL; jar = jar->next) {
            if (CoffeeIsJar(jar) && CoffeeIsClosed(jar)) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if (CoffeeIsHand(hand) && CoffeeHolds(hand, jar)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    else if (CoffeeIsHand(effector)) {
        for (jar = objects; jar != NULL; jar = jar->next) {
            if (CoffeeHolds(effector, jar) && CoffeeIsJar(jar) && CoffeeIsClosed(jar)) {
                for (hand = resources; hand != NULL; hand = hand->next) {
                    if (CoffeeIsEmptyHand(hand)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN screw_closed_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *jar, *lid;
    CsResource *other_hand;

    if (CoffeeIsHand(effector)) {
        for (jar = objects; jar != NULL; jar = jar->next) {
            if (CoffeeIsJar(jar) && CoffeeIsOpen(jar)) {
                for (lid = objects; lid != NULL; lid = lid->next) {
                    if (CoffeeIsLidOfJar(lid, jar)) {
                        for (other_hand = resources; other_hand != NULL; other_hand = other_hand->next) {
                            if (CoffeeIsHand(other_hand) && CoffeeHolds(effector, jar) && CoffeeHolds(other_hand, lid)) {
                                return(TRUE);
                            }
                            else if (CoffeeIsHand(other_hand) && CoffeeHolds(effector, lid) && CoffeeHolds(other_hand, jar)) {
                                return(TRUE);
                            }
                        }
                    }
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN swap_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *object;
    CsResource *other_hand;

    if (CoffeeIsEmptyHand(effector)) {
        for (object = objects; object != NULL; object = object->next) {
            for (other_hand = resources; other_hand != NULL; other_hand = other_hand->next) {
                if (CoffeeIsHand(other_hand) && CoffeeHolds(other_hand, object)) {
                    return(TRUE);
                }
            }
        }
    }
    else {
        for (object = objects; object != NULL; object = object->next) {
            if (CoffeeHolds(effector, object)) {
                for (other_hand = resources; other_hand != NULL; other_hand = other_hand->next) {
                    if (CoffeeIsEmptyHand(other_hand)) {
                        return(TRUE);
                    }
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN drink_triggers_effector(CsResource *effector, int cycle)
{
    CsObject *object;

    if (CoffeeIsHand(effector)) {
        for (object = objects; object != NULL; object = object->next) {
            if (CoffeeIsNonemptyContainer(object)) {
                return(TRUE);
            }
        }
    }
    return(FALSE);
}

double CoffeeSituationTriggersObject(int cycle, ArgumentFunction f)
{
    return(0.0);
}

/******** END OF FILE *********************************************************/
