#include "cs.h"

extern void EnvPreActionTrace(int cycle, char *name, GENERIC_LIST *features);
extern void EnvPostActionTrace(GENERIC_LIST *features);
extern void CsMessage(MessageType category, int cycle, char *buffer);
extern void CsActionFailureMessage(int cycle, char *reason);
extern void CsArgumentFailureMessage(int cycle, char *action);
extern void ActRecord(ActionType act, int cycle, char *arg1, char *arg2);

extern TERM *get_feature(GENERIC_LIST *features, char *functor);

/******************************************************************************/

static void mix_list(TERM *list)
{
    if ((list != NULL) && (list->sort == LIST)) {
        GENERIC_LIST *tmp;
        char contents[20];
        int i;

        for (tmp = list->arguments, i = 0; ((tmp != NULL) && (i < 20)); tmp = tmp->next, i++) {
            contents[i] = ((char *)(tmp->head))[0];
        }
        for (tmp = list->arguments; (tmp != NULL); tmp = tmp->next) {
            tmp->head = copy_string(contents);
        }
    }
    else {
        fprintf(stdout, "WARNING: Argument to mix_list is not a list\n");
    }
}

static CsObject *first_element(TERM *list)
{
    BOOLEAN found = FALSE;
    CsObject *tmp = objects;

    if ((list != NULL) && (list->sort == LIST)) {
        while ((tmp != NULL) && !found) {
            found = (strcmp(tmp->name, ((char *)(list->arguments->head))) == 0);
            if (!found) {
                tmp = tmp->next;
            }
        }
        return(tmp);
    }
    else {
        fprintf(stdout, "WARNING: Argument to first_element is not a list\n");
        return(NULL);
    }
}

static BOOLEAN empty_list(TERM *list)
{
    if ((list != NULL) && (list->sort == LIST)) {
        return(list->arguments == NULL);
    }
    else {
        fprintf(stdout, "WARNING: Argument to empty_list is not a list\n");
        return(FALSE);
    }
}

static BOOLEAN member(char *name, TERM *list)
{
    BOOLEAN found = FALSE;
    GENERIC_LIST *tmp;

    if ((list != NULL) && (list->sort == LIST)) {
        for (tmp = list->arguments; ((tmp != NULL) && !found); tmp = tmp->next) {
            found = (strcmp(name, ((char *)(tmp->head))) == 0);
        }
    }
    else {
        fprintf(stdout, "WARNING: Second argument to member is not a list\n");
    }

    return(found);
}

/******************************************************************************/

BOOLEAN CoffeeHolds(CsResource *effector, CsObject *object)
{
    TERM *holding = get_feature(effector->features, "holds");

    if (holding != NULL) {
        return( (member(object->name, holding->arguments->head)) );
    }
    else {
        return(FALSE);
    }
}

BOOLEAN CoffeeIsContainer(CsObject *object)
{
    return(get_feature(object->features, "contains") != NULL);
}

BOOLEAN CoffeeContains(CsObject *object, char *functor)
{
    TERM *contents = get_feature(object->features, "contains");

    if ((contents != NULL) && (contents->arguments != NULL)) {
        return(member(functor, contents->arguments->head));
    }
    else {
        return(FALSE);
    }
}

BOOLEAN CoffeeIsEmptyContainer(CsObject *source)
{
    TERM *functor = get_feature(source->features, "contains");

    return((functor != NULL) && ((TERM *)(functor->arguments->head))->arguments == NULL);
}

BOOLEAN CoffeeIsNonemptyContainer(CsObject *source)
{
    TERM *functor = get_feature(source->features, "contains");

    return((functor != NULL) && ((TERM *)(functor->arguments->head))->arguments != NULL);
}

BOOLEAN CoffeeIsSource(CsObject *object)
{
    return(CoffeeIsContainer(object) && !CoffeeIsEmptyContainer(object));
}

BOOLEAN CoffeeIsStirrer(CsObject *object)
{
    return((strcmp(object->name, "teaspoon") == 0) ||
           (strcmp(object->name, "stirrer") == 0));
}

BOOLEAN CoffeeIsSpoon(CsObject *object)
{
    return(strcmp(object->name, "teaspoon") == 0);
}

BOOLEAN CoffeeIsBowl(CsObject *object)
{
    return((strcmp(object->name, "sugar bowl") == 0) ||
           (strcmp(object->name, "cereal bowl") == 0));
}

BOOLEAN CoffeeIsJar(CsObject *object)
{
    return(strcmp(object->name, "coffee jar") == 0);
}

BOOLEAN CoffeeIsMug(CsObject *object)
{
    return(strcmp(object->name, "coffee mug") == 0);
}

BOOLEAN CoffeeIsPacket(CsObject *object)
{
    return((strcmp(object->name, "sugar packet") == 0) ||
           (strcmp(object->name, "coffee packet") == 0) ||
           (strcmp(object->name, "salt packet") == 0));
}

BOOLEAN CoffeeIsCarton(CsObject *object)
{
    return(strcmp(object->name, "milk carton") == 0);
}

BOOLEAN CoffeeIsLid(CsObject *object)
{
    return(strcmp(object->name, "coffee jar lid") == 0);
}

BOOLEAN CoffeeIsOpen(CsObject *object)
{
    return(get_feature(object->features, "open") != NULL);
}

BOOLEAN CoffeeIsClosed(CsObject *object)
{
    return(get_feature(object->features, "closed") != NULL);
}

BOOLEAN CoffeeIsHeld(CsObject *object)
{
    TERM *functor = get_feature(object->features, "position");

    return((functor != NULL) && (strcmp(((TERM *)(functor->arguments->head))->functor, "held_by") == 0));
}

BOOLEAN CoffeeIsHand(CsResource *effector)
{
    return(TRUE);
}

CsResource *CoffeeGetOtherHand(CsResource *hand)
{
    CsResource *tmp;

    for (tmp = resources; (tmp != NULL); tmp = tmp->next) {
        if (CoffeeIsHand(tmp) && (tmp != hand)) {
            return(hand);
        }
    }
    return(NULL);
}

BOOLEAN CoffeeIsEmptyHand(CsResource *effector)
{
    TERM *functor = get_feature(effector->features, "holds");

    return( CoffeeIsHand(effector) && (((TERM *)(functor->arguments->head))->arguments == NULL) );
}

BOOLEAN CoffeeIsLidOfJar(CsObject *lid, CsObject *jar)
{
    return(CoffeeIsLid(lid) && CoffeeIsJar(jar));
}

BOOLEAN CoffeeNotHeldBySameHand(CsObject *object1, CsObject *object2)
{
    CsResource *tmp;

    for (tmp = resources; (tmp != NULL); tmp = tmp->next) {
        if (CoffeeIsHand(tmp) && CoffeeHolds(tmp, object1) && CoffeeHolds(tmp, object2)) {
            return(FALSE);
        }
    }
    return(TRUE);
}

static BOOLEAN stoi(char *string, int *intptr)
{
    BOOLEAN error = FALSE;
    int sign = 1, n = 0;

    if (*string == '+') {
        sign = 1;
        string++;
    }
    else if (*string == '-') {
        sign = -1;
        string++;
    }

    while ((*string != '\0') && !error) {
        int j = (*string - '0');
        j = (*string - '0');
        error = (j > 9) || (j < 0);
        n = 10 * n + j;
        string++;
    }

    *intptr = sign * n;

    return(!error);
}

BOOLEAN CoffeeHandReaches(CsResource *hand, CsObject *object)
{
    int x, y;
    TERM *position = get_feature(object->features, "position");

    if ((position != NULL) && (position->sort == GROUND) && (position->arguments != NULL)) {
        if ((position->arguments->next == NULL) || (position->arguments->next->next != NULL)) {
            return(FALSE);
        }
        else if (stoi(((TERM *)(position->arguments->head))->functor, &x) &&
                 stoi(((TERM *)(position->arguments->next->head))->functor, &y)) {

            /* CoffeeHandReaches(Left Hand, position(X,Y))  :- Y < (4-X).      */

            if (strcmp(hand->name, "Left Hand") == 0) {
                return((y < (4-x)));
            }

            /* CoffeeHandReaches(Right Hand, position(X,Y)) :- Y < (X+4).      */
            else if (strcmp(hand->name, "Right Hand") == 0) {
                return((y < (x+4)));
            }
            else {
                return(FALSE);
            }
        }
        else {
            return(FALSE);
        }
    }
    else {
        return(FALSE);
    }
}

BOOLEAN CoffeeIsWithinReach(CsObject *object)
{
    CsResource *hand;

    for (hand = resources; hand != NULL; hand = hand->next) {
        if (CoffeeIsEmptyHand(hand) && CoffeeHandReaches(hand, object)) {
            return(TRUE);
        }
    }
    return(FALSE);
}

BOOLEAN CoffeeItemsIncludeSpoon(TERM *list)
{
    CsObject *spoon;

    if (!empty_list(list)) {
        for (spoon = objects; (spoon != NULL); spoon = spoon->next) {
            if (CoffeeIsSpoon(spoon) && member(spoon->name, list)) {
                return(TRUE);
            }
        }
    }
    return(FALSE);
}

BOOLEAN CoffeeItemsIncludeStirrer(TERM *list)
{
    CsObject *stirrer;

    if (!empty_list(list)) {
        for (stirrer = objects; (stirrer != NULL); stirrer = stirrer->next) {
            if (CoffeeIsStirrer(stirrer) && member(stirrer->name, list)) {
                return(TRUE);
            }
        }
    }
    return(FALSE);
}

BOOLEAN CoffeeItemsIncludeOpenSource(TERM *list)
{
    CsObject *source;

    if (!empty_list(list)) {
        for (source = objects; source != NULL; source = source->next) {
            if (CoffeeIsSource(source) && CoffeeIsOpen(source) && member(source->name, list)) {
                return(TRUE);
            }
        }
    }
    return(FALSE);
}

BOOLEAN CoffeeItemsIncludeScrewableLid(TERM *list)
{
    CsObject *lid;

    if (!empty_list(list)) {
        for (lid = objects; lid != NULL; lid = lid->next) {
            if (CoffeeIsLid(lid) && member(lid->name, list)) {
                return(TRUE);
            }
        }
    }
    return(FALSE);
}

BOOLEAN CoffeeItemsIncludeUnscrewableContainer(TERM *list)
{
    CsObject *container;

    if (!empty_list(list)) {
        for (container = objects; container != NULL; container = container->next) {
            if (CoffeeIsJar(container) && CoffeeIsClosed(container) && member(container->name, list)) {
                return(TRUE);
            }
        }
    }
    return(FALSE);
}

BOOLEAN CoffeeItemsIncludeScrewableContainer(TERM *list)
{
    CsObject *container;

    if (!empty_list(list)) {
        for (container = objects; container != NULL; container = container->next) {
            if (CoffeeIsJar(container) && CoffeeIsOpen(container) && member(container->name, list)) {
                return(TRUE);
            }
	}
    }
    return(FALSE);
}

BOOLEAN CoffeeItemsIncludeTearableContainer(TERM *list)
{
    CsObject *container;

    if (!empty_list(list)) {
        for (container = objects; container != NULL; container = container->next) {
            if ((CoffeeIsPacket(container) || CoffeeIsCarton(container)) && member(container->name, list)) {
                return(TRUE);
            }
        }
    }
    return(FALSE);
}

TERM *CoffeeContentsOfHand(CsResource *hand)
{
    TERM *held = get_feature(hand->features, "holds");

    if (held == NULL) {
        return(NULL);
    }
    else {
        return((TERM *)(held->arguments->head));
    }
}

/******************************************************************************/

static void pre_append(TERM *functor, char *element)
{
    GENERIC_LIST *tmp;

    if ((tmp = (GENERIC_LIST *)malloc(sizeof(GENERIC_LIST))) == NULL) {
        fprintf(stderr, "WARNING: Insufficient memory for malloc");
    }
    else {
        /* Add the new list element as the first element of the list */

        tmp->head = copy_string(element);
        tmp->next = functor->arguments;
        functor->arguments = tmp;
    }
}

/******************************************************************************/
/********* pick_up(ObjectWithinReach, EmptyHand) ******************************/

void pick_up(int cycle, CsObject *argument, CsResource *effector)
{
    if ((argument == NULL) || (effector == NULL)) {
        CsArgumentFailureMessage(cycle, "pick_up");
    }
    else {
        char buf[80];

        sprintf(buf, "        Picking up %s with %s", argument->name, effector->name);
        CsMessage(ACTION, cycle, buf);

        if (!CoffeeIsEmptyHand(effector)) {
            CsActionFailureMessage(cycle, "Hand is full");
        }
        else if (!CoffeeHandReaches(effector, argument)) {
            CsActionFailureMessage(cycle, "Object is out of reach");
        }
        else {
            TERM *functor;
 
            EnvPreActionTrace(cycle, effector->name, effector->features);
            functor = get_feature(effector->features, "holds");
            pre_append(((TERM *)(functor->arguments->head)), argument->name);
            EnvPostActionTrace(effector->features);

            EnvPreActionTrace(cycle, argument->name, argument->features);
            functor = get_feature(argument->features, "position");
            ((TERM *)(functor->arguments->head))->functor = copy_string("held_by");
            ((TERM *)(functor->arguments->head))->arguments = functor->arguments->next;
            functor->arguments->next = NULL;
            ((TERM *)(((TERM *)(functor->arguments->head))->arguments->head))->functor = copy_string(effector->name);
            EnvPostActionTrace(argument->features);

            ActRecord(PICK_UP, cycle, argument->name, NULL);
        }
    }
}

/******************************************************************************/
/********* put_down(HandContainingObject) *************************************/

void put_down(int cycle, CsResource *effector)
{
    if (effector == NULL) {
        CsArgumentFailureMessage(cycle, "put_down");
    }
    else {

        char buf[80];
        TERM *functor;
        GENERIC_LIST *tmp;

        functor = get_feature(effector->features, "holds");
        for (tmp = ((TERM *)(functor->arguments->head))->arguments; tmp != NULL; tmp = tmp->next) {
            sprintf(buf, "        Putting down %s", (char *)(tmp->head));
            CsMessage(ACTION, cycle, buf);
        }

        for (tmp = ((TERM *)(functor->arguments->head))->arguments; tmp != NULL; tmp = tmp->next) {
            ActRecord(PUT_DOWN, cycle, (char *)(tmp->head), NULL);
        }

        if (CoffeeIsEmptyHand(effector)) {
            CsActionFailureMessage(cycle, "Hand is empty");
        }
        else {
            tmp = ((TERM *)(functor->arguments->head))->arguments;

            EnvPreActionTrace(cycle, effector->name, effector->features);
            ((TERM *)(functor->arguments->head))->arguments = NULL;
            EnvPostActionTrace(effector->features);

            while (tmp != NULL) {
                CsObject *tmp2 = objects;

                while ((tmp2 != NULL) && (strcmp(tmp2->name, (char *)(tmp->head)) != 0)) {
                    tmp2 = tmp2->next;
                }
                if (tmp2 != NULL) {
                    EnvPreActionTrace(cycle, tmp2->name, tmp2->features);

                    functor = get_feature(tmp2->features, "position");
                    if (strcmp(effector->name, "Left Hand") == 0) {
                        ((TERM *)(functor->arguments->head))->functor = copy_string("-1");
                    }
                    else {
                        ((TERM *)(functor->arguments->head))->functor = copy_string("1");
                    }
                    functor->arguments->next = ((TERM *)(functor->arguments->head))->arguments;
                    ((TERM *)(functor->arguments->head))->arguments = NULL;
                    ((TERM *)(functor->arguments->next->head))->functor = copy_string("1");

                    EnvPostActionTrace(tmp2->features);
                }
                tmp = tmp->next;
            }
        }
    }
}

/******************************************************************************/
/********* dip_spoon(HandContainingSpoon, OpenSource) *************************/

void dip_spoon(int cycle, CsObject *argument, CsResource *effector)
{
    if ((argument == NULL) || (effector == NULL)) {
        CsArgumentFailureMessage(cycle, "dip_spoon");
    }
    else {

        char buf[80];
        CsObject *spoon = NULL;

        if (!CoffeeIsEmptyHand(effector)) {
            spoon = first_element(CoffeeContentsOfHand(effector));
            sprintf(buf, "        Dipping %s into %s", spoon->name, argument->name);
            CsMessage(ACTION, cycle, buf);
        }

        if (CoffeeIsEmptyHand(effector)) {
            CsActionFailureMessage(cycle, "Hand is empty");
        }
        else if (!CoffeeIsSpoon(spoon)) {
            CsActionFailureMessage(cycle, "Hand doesn't hold a spoon");
        }
        else if (!CoffeeIsEmptyContainer(spoon)) {
            CsActionFailureMessage(cycle, "Spoon isn't empty");
        }
        else if (!CoffeeIsOpen(argument)) {
            CsActionFailureMessage(cycle, "Source is not open");
        }
        else if (CoffeeIsEmptyContainer(argument)) {
            CsActionFailureMessage(cycle, "Source is empty");
        }
        else if (!CoffeeIsContainer(argument)) {
            CsActionFailureMessage(cycle, "Source is not a container");
        }
        else {
            TERM *functor;
            GENERIC_LIST *tmp;

            EnvPreActionTrace(cycle, argument->name, argument->features);
            functor = get_feature(argument->features, "contains");
            tmp = ((TERM *)(functor->arguments->head))->arguments;
            ((TERM *)(functor->arguments->head))->arguments = tmp->next;
            EnvPostActionTrace(argument->features);

            EnvPreActionTrace(cycle, spoon->name, spoon->features);
            functor = get_feature(spoon->features, "contains");
            tmp->next = ((TERM *)(functor->arguments->head))->arguments;
            ((TERM *)(functor->arguments->head))->arguments = tmp;
            EnvPostActionTrace(spoon->features);

            ActRecord(DIP_SPOON, cycle, argument->name, NULL);
        }
    }
}

/******************************************************************************/
/********* empty_spoon(HandContainingSpoon, OpenContainer) ********************/

void empty_spoon(int cycle, CsObject *argument, CsResource *effector)
{
    if ((argument == NULL) || (effector == NULL)) {
        CsArgumentFailureMessage(cycle, "empty_spoon");
    }
    else {

        char buf[80];
        CsObject *spoon = NULL;

        if (!CoffeeIsEmptyHand(effector)) {
            spoon = first_element(CoffeeContentsOfHand(effector));

            sprintf(buf, "        Emptying %s into %s", spoon->name, argument->name);
            CsMessage(ACTION, cycle, buf);
        }

        if (CoffeeIsEmptyHand(effector)) {
            CsActionFailureMessage(cycle, "Hand is empty");
        }
        else if (!CoffeeIsSpoon(spoon)) {
            CsActionFailureMessage(cycle, "Hand doesn't hold a spoon");
        }
        else if (CoffeeIsEmptyContainer(spoon)) {
            CsActionFailureMessage(cycle, "Spoon is empty");
        }
        else if (!CoffeeIsOpen(argument)) {
            CsActionFailureMessage(cycle, "Target is not open");
        }
        else if (!CoffeeIsContainer(argument)) {
            CsActionFailureMessage(cycle, "Target is not a container");
        }
        else {

            TERM *functor;
            GENERIC_LIST *tmp;

            EnvPreActionTrace(cycle, spoon->name, spoon->features);
            functor = get_feature(spoon->features, "contains");
            tmp = ((TERM *)(functor->arguments->head))->arguments;
            ((TERM *)(functor->arguments->head))->arguments = tmp->next;
            EnvPostActionTrace(spoon->features);

            EnvPreActionTrace(cycle, argument->name, argument->features);
            functor = get_feature(argument->features, "contains");
            tmp->next = ((TERM *)(functor->arguments->head))->arguments;
            ((TERM *)(functor->arguments->head))->arguments = tmp;
            EnvPostActionTrace(argument->features);

            ActRecord(EMPTY_SPOON, cycle, (char *)(tmp->head), argument->name);
        }
    }
}

/******************************************************************************/

void consume_spoon_contents(int cycle, CsResource *effector)
{
    if (effector == NULL) {
        CsArgumentFailureMessage(cycle, "consume_spoon_contents");
    }
    else {

        char buf[80];
        CsObject *spoon = NULL;

        if (!CoffeeIsEmptyHand(effector)) {
            spoon = first_element(CoffeeContentsOfHand(effector));

            sprintf(buf, "        Emptying %s into mouth", spoon->name);
            CsMessage(ACTION, cycle, buf);
        }

        if (CoffeeIsEmptyHand(effector)) {
            CsActionFailureMessage(cycle, "Hand is empty");
        }
        else if (!CoffeeIsSpoon(spoon)) {
            CsActionFailureMessage(cycle, "Hand doesn't hold a spoon");
        }
        else if (CoffeeIsEmptyContainer(spoon)) {
            CsActionFailureMessage(cycle, "Spoon is empty");
        }
        else {

            TERM *functor;
            GENERIC_LIST *tmp;

            EnvPreActionTrace(cycle, spoon->name, spoon->features);
            functor = get_feature(spoon->features, "contains");
            tmp = ((TERM *)(functor->arguments->head))->arguments;
            ((TERM *)(functor->arguments->head))->arguments = tmp->next;
            EnvPostActionTrace(spoon->features);

            ActRecord(EAT_FROM_SPOON, cycle, (char *)(tmp->head), NULL);
        }
    }
}

/******************************************************************************/
/********* stir(HandContainingStirrer, NonEmptyOpenContainer) *****************/

void stir(int cycle, CsObject *argument, CsResource *effector)
{
    if ((argument == NULL) || (effector == NULL)) {
        CsArgumentFailureMessage(cycle, "stir");
    }
    else {
        char buf[80];
        CsObject *stirrer = NULL;

        if (!CoffeeIsEmptyHand(effector)) {
            stirrer = first_element(CoffeeContentsOfHand(effector));

            sprintf(buf, "        Stirring %s with %s", argument->name, stirrer->name);
            CsMessage(ACTION, cycle, buf);
        }

        if (CoffeeIsEmptyHand(effector)) {
            CsActionFailureMessage(cycle, "Hand is empty");
        }
        else if (!CoffeeIsStirrer(stirrer)) {
            CsActionFailureMessage(cycle, "Hand doesn't hold a stirrer");
        }
        else if (!CoffeeIsContainer(argument)) {
            CsActionFailureMessage(cycle, "Object to be stirred is not a container");
        }
        else if (CoffeeIsEmptyContainer(argument)) {
            CsActionFailureMessage(cycle, "Object to be stirred is empty");
        }
        else if (!CoffeeIsOpen(argument)) {
            CsActionFailureMessage(cycle, "Object to be stirred is not open");
        }
        else {
            TERM *functor;

            EnvPreActionTrace(cycle, argument->name, argument->features);
            functor = get_feature(argument->features, "contains");
            mix_list((TERM *)(functor->arguments->head));
            EnvPostActionTrace(argument->features);

            ActRecord(STIR, cycle, argument->name, NULL);
	}
    }
}

/******************************************************************************/
/********* pour(HandContainingOpenSource, OpenContainer) **********************/

void pour(int cycle, CsObject *argument, CsResource *effector)
{
    if (argument == NULL) {
        CsArgumentFailureMessage(cycle, "pour (no suitable target)");
    }
    else if (effector == NULL) {
        CsArgumentFailureMessage(cycle, "pour (no suitable effector)");
    }
    else {
        char buf[80];
        CsObject *source = NULL;

        if (!CoffeeIsEmptyHand(effector)) {
            source = first_element(CoffeeContentsOfHand(effector));

            sprintf(buf, "        Pouring %s into %s", source->name, argument->name);
            CsMessage(ACTION, cycle, buf);
        }

        if (CoffeeIsEmptyHand(effector)) {
            CsActionFailureMessage(cycle, "Hand is empty");
        }
        else if (!CoffeeIsOpen(source)) {
            CsActionFailureMessage(cycle, "Source is not open");
        }
        else if (CoffeeIsEmptyContainer(source)) {
            CsActionFailureMessage(cycle, "Source is empty");
        }
        else if (!CoffeeIsContainer(argument)) {
            CsActionFailureMessage(cycle, "Target is not a container");
        }
        else if (!CoffeeIsOpen(argument)) {
            CsActionFailureMessage(cycle, "Target is not open");
        }
        else if (CoffeeHolds(effector, argument)) {
            CsActionFailureMessage(cycle, "Cannot pour into self");
        }
        else {
            TERM *functor;
            GENERIC_LIST *tmp;

            EnvPreActionTrace(cycle, source->name, source->features);
            functor = get_feature(source->features, "contains");
            tmp = ((TERM *)(functor->arguments->head))->arguments;
            ((TERM *)(functor->arguments->head))->arguments = tmp->next;
            EnvPostActionTrace(source->features);

            EnvPreActionTrace(cycle, argument->name, argument->features);
            functor = get_feature(argument->features, "contains");
            tmp->next = ((TERM *)(functor->arguments->head))->arguments;
            ((TERM *)(functor->arguments->head))->arguments = tmp;
            EnvPostActionTrace(argument->features);

            ActRecord(POUR, cycle, (char *)(tmp->head), argument->name);
        }
    }
}

/******************************************************************************/
/********* tear(HandContainingPacket, EmptyHand) ******************************/

void tear(int cycle, CsResource *effector1, CsResource *effector2)
{
    if ((effector1 == NULL) || (effector2 == NULL)) {
        CsArgumentFailureMessage(cycle, "tear");
    }
    else {

        char buf[80];
        CsObject *pack = NULL;

        if (!CoffeeIsEmptyHand(effector1)) {

            pack = first_element(CoffeeContentsOfHand(effector1));

            sprintf(buf, "        Tearing %s (with %s)", pack->name, effector2->name);
            CsMessage(ACTION, cycle, buf);
        }

        if (CoffeeIsEmptyHand(effector1)) {
            CsActionFailureMessage(cycle, "Hand is empty");
        }
        else if (!(CoffeeIsPacket(pack) || CoffeeIsCarton(pack))) {
            CsActionFailureMessage(cycle, "Holding hand doesn't contain a packet or carton");
        }
        else if (CoffeeIsOpen(pack)) {
            CsActionFailureMessage(cycle, "Packet/Carton is already open");
        }
        else if (!CoffeeIsEmptyHand(effector2)) {
            CsActionFailureMessage(cycle, "Tearing hand is not empty");
        }
        else {
            TERM *functor;

            EnvPreActionTrace(cycle, pack->name, pack->features);
            functor = get_feature(pack->features, "closed");
            functor->functor = copy_string("open");
            EnvPostActionTrace(pack->features);

            ActRecord(TEAR, cycle, pack->name, NULL);
        }
    }
}

/******************************************************************************/
/********* screw_open(HandContainingJar, EmptyHand) ***************************/

void screw_open(int cycle, CsResource *effector1, CsResource *effector2)
{
    if ((effector1 == NULL) || (effector2 == NULL)) {
        CsArgumentFailureMessage(cycle, "screw_open");
    }
    else {

        char buf[80];
        CsObject *jar = NULL;

        if (!CoffeeIsEmptyHand(effector1)) {
            jar = first_element(CoffeeContentsOfHand(effector1));

            sprintf(buf, "        Screwing open %s (with %s)", jar->name, effector2->name);
            CsMessage(ACTION, cycle, buf);
        }

        if (CoffeeIsEmptyHand(effector1)) {
            CsActionFailureMessage(cycle, "First hand is empty");
        }
        else if (!(CoffeeIsJar(jar))) {
            CsActionFailureMessage(cycle, "Holding hand doesn't contain a jar");
        }
        else if (CoffeeIsOpen(jar)) {
            CsActionFailureMessage(cycle, "Jar is already open");
        }
        else if (!CoffeeIsEmptyHand(effector2)) {
            CsActionFailureMessage(cycle, "Second hand is not empty");
        }
        else {
            TERM *functor;
            CsObject *lid = NULL;
            CsObject *tmp = objects;

            while ((tmp != NULL) && (lid == NULL)) {
                functor = get_feature(tmp->features, "position");
                if ((functor != NULL) && (functor->arguments != NULL)) {
                    functor = ((TERM *)(functor->arguments->head));
                    if ((strcmp(functor->functor, "on_jar") == 0) && (functor != NULL) && (functor->arguments != NULL)) {
                        functor = ((TERM *)(functor->arguments->head));
                        if (strcmp(functor->functor, jar->name) == 0) {
                            lid = tmp;
                        }
                    }
                }
            }

            EnvPreActionTrace(cycle, jar->name, jar->features);
            functor = get_feature(jar->features, "closed");
            functor->functor = copy_string("open");
            EnvPostActionTrace(jar->features);

            EnvPreActionTrace(cycle, effector2->name, effector2->features);
            functor = get_feature(effector2->features, "holds");
            pre_append(((TERM *)(functor->arguments->head)), lid->name);
            EnvPostActionTrace(effector2->features);

            if (lid == NULL) {
                fprintf(stdout, "HELP: I can't find the lid of %s", jar->name);
            }
            else {
                EnvPreActionTrace(cycle, lid->name, lid->features);
                functor = get_feature(lid->features, "position");
                functor = ((TERM *)(functor->arguments->head));
                functor->functor = copy_string("held_by");
                functor = ((TERM *)(functor->arguments->head));
                functor->functor = copy_string(effector2->name);
                EnvPostActionTrace(lid->features);
            }
            ActRecord(SCREW_OPEN, cycle, jar->name, NULL);
        }
    }
}

/******************************************************************************/
/********* screw_closed(HandContainingJar, HandContainingScrewLid) ************/

void screw_closed(int cycle, CsResource *effector1, CsResource *effector2)
{
    if ((effector1 == NULL) || (effector2 == NULL)) {
        CsArgumentFailureMessage(cycle, "screw_closed_up");
    }
    else {

        char buf[80];
        CsObject *jar = NULL, *lid = NULL;

        if (!CoffeeIsEmptyHand(effector1) && !CoffeeIsEmptyHand(effector2)) {
            jar = first_element(CoffeeContentsOfHand(effector1));
            lid = first_element(CoffeeContentsOfHand(effector2));

            sprintf(buf, "        Screwing closed %s (with %s)", jar->name, effector2->name);
            CsMessage(ACTION, cycle, buf);
        }

        if (CoffeeIsEmptyHand(effector1)) {
            CsActionFailureMessage(cycle, "First hand is empty");
        }
        else if (!(CoffeeIsJar(jar))) {
            CsActionFailureMessage(cycle, "Hand doesn't contain jar");
        }
        else if (CoffeeIsClosed(jar)) {
            CsActionFailureMessage(cycle, "Jar is already closed");
        }
        else if (CoffeeIsEmptyHand(effector2)) {
            CsActionFailureMessage(cycle, "Second hand is empty");
        }
        else if (!CoffeeIsLidOfJar(lid, jar)) {
            CsActionFailureMessage(cycle, "Second hand doesn't hold lid of jar");
        }
        else {
            TERM *functor;

            EnvPreActionTrace(cycle, jar->name, jar->features);
            functor = get_feature(jar->features, "open");
            functor->functor = copy_string("closed");
            EnvPostActionTrace(jar->features);

            EnvPreActionTrace(cycle, lid->name, lid->features);
            functor = get_feature(lid->features, "position");
            functor = ((TERM *)(functor->arguments->head));
            functor->functor = copy_string("on_jar");
            functor = ((TERM *)(functor->arguments->head));
            functor->functor = copy_string(jar->name);
            EnvPostActionTrace(lid->features);

            EnvPreActionTrace(cycle, effector2->name, effector2->features);
            functor = get_feature(effector2->features, "holds");
            functor = ((TERM *)(functor->arguments->head));
            free(functor->arguments);
            functor->arguments = NULL;
            EnvPostActionTrace(effector2->features);

            ActRecord(SCREW_CLOSED, cycle, jar->name, NULL);
	}
    }
}

/******************************************************************************/
/********* swap(HandContainingObject, EmptyHand) ******************************/

void swap(int cycle, CsResource *effector1, CsResource *effector2)
{
    if ((effector1 == NULL) || (effector2 == NULL)) {
        CsArgumentFailureMessage(cycle, "swap");
    }
    else {

        char buf[80];
        CsObject *object = NULL;

        if (!CoffeeIsEmptyHand(effector1)) {
            object = first_element(CoffeeContentsOfHand(effector1));

            sprintf(buf, "        Swapping %s from %s to %s", object->name, effector1->name, effector2->name);
            CsMessage(ACTION, cycle, buf);
        }

        if (!CoffeeIsEmptyHand(effector2)) {
            CsActionFailureMessage(cycle, "Target hand is full");
        }
        else if (CoffeeIsEmptyHand(effector1)) {
            CsActionFailureMessage(cycle, "Source hand is empty");
        }
        else {
            TERM *functor;
            GENERIC_LIST *tmp;

            EnvPreActionTrace(cycle, effector1->name, effector1->features);
            functor = get_feature(effector1->features, "holds");
            tmp = ((TERM *)(functor->arguments->head))->arguments;
            ((TERM *)(functor->arguments->head))->arguments = tmp->next;
            EnvPostActionTrace(effector1->features);

            EnvPreActionTrace(cycle, effector2->name, effector2->features);
            functor = get_feature(effector2->features, "holds");
            tmp->next = ((TERM *)(functor->arguments->head))->arguments;
            ((TERM *)(functor->arguments->head))->arguments = tmp;
            EnvPostActionTrace(effector2->features);

            EnvPreActionTrace(cycle, object->name, object->features);
            functor = get_feature(object->features, "position");
            ((TERM *)(functor->arguments->head))->functor = copy_string(effector2->name);
            EnvPostActionTrace(object->features);

            ActRecord(SWAP, cycle, NULL, NULL);
        }
    }
}

/******************************************************************************/
/********* drink(HandContainingOpenSource) **********************/

void drink(int cycle, CsResource *effector)
{
    if (effector == NULL) {
        CsArgumentFailureMessage(cycle, "drink (no suitable effector)");
    }
    else {
        char buf[80];
        CsObject *source = NULL;

        if (!CoffeeIsEmptyHand(effector)) {
            source = first_element(CoffeeContentsOfHand(effector));

            sprintf(buf, "        Drinking from %s", source->name);
            CsMessage(ACTION, cycle, buf);
        }

        if (CoffeeIsEmptyHand(effector)) {
            CsActionFailureMessage(cycle, "Hand is empty");
        }
        else if (!CoffeeIsOpen(source)) {
            CsActionFailureMessage(cycle, "Source is not open");
        }
        else if (CoffeeIsEmptyContainer(source)) {
            CsActionFailureMessage(cycle, "Source is empty");
        }
        else {
            TERM *functor;
            GENERIC_LIST *tmp;

            EnvPreActionTrace(cycle, source->name, source->features);
            functor = get_feature(source->features, "contains");
            tmp = ((TERM *)(functor->arguments->head))->arguments;
            ((TERM *)(functor->arguments->head))->arguments = tmp->next;
            EnvPostActionTrace(source->features);

            ActRecord(DRINK, cycle, source->name, NULL);
        }
    }
}

/******************************************************************************/

BOOLEAN CoffeeStepWorld(int cycle)
{
    extern BOOLEAN goal_is_achieved_on_current_cycle(char *goal);
    extern char *terminating_goal;

    return(goal_is_achieved_on_current_cycle(terminating_goal));
}

/******************************************************************************/
