/*******************************************************************************

  This file contains definitions of objects and the triggers between schema,
  objects, and effectors.

*******************************************************************************/

#include "cs.h"

extern CsObject *CsInstallObjectName(GENERIC_LIST **objects, char *name);
extern TERM *CsInstallObjectFeatureFunctor(CsObject *object, char *functor);
extern CsResource *CsInstallResourceName(CsResource **effectors, char *name);
extern TERM *CsInstallResourceFeatureFunctor(CsResource *effector, char *functor);
extern void CsNormaliseObjects(GENERIC_LIST *objects, int cycle);
extern void CsNormaliseResources(CsResource *effectors, int cycle);

extern CsObject *CsGetObject(GENERIC_LIST *object_list, char *object_name);
extern void CsResetObjectFeatures(CsObject *object);
extern CsResource *CsGetResource(CsResource *effector_list, char *effector_name);
extern void CsResetResourceFeatures(CsResource *effector);
extern TERM *install_argument_functor(TERM *functor, char *name);
extern void install_list_argument(TERM *functor, int n, ...);

extern double RtSituationTriggersButton(int cycle, ArgumentFunction f);
extern double RtSituationTriggersVisualObject(int cycle, ArgumentFunction f);
extern double RtSituationTriggersAuditoryObject(int cycle, ArgumentFunction f);

/******************************************************************************/

GENERIC_LIST *ColavitaInitialiseObjects()
{
    GENERIC_LIST *objects = NULL;

    CsObject *object;

    object = CsInstallObjectName(&objects, "auditory");
    object->triggering_situation = *RtSituationTriggersAuditoryObject;
    object = CsInstallObjectName(&objects, "visual");
    object->triggering_situation = *RtSituationTriggersVisualObject;
    object = CsInstallObjectName(&objects, "l_button");
    object->triggering_situation = *RtSituationTriggersButton;
    object = CsInstallObjectName(&objects, "r_button");
    object->triggering_situation = *RtSituationTriggersButton;

    return(objects);

}

/******************************************************************************/

CsResource *ColavitaInitialiseResources()
{
    CsResource *effectors = NULL;

    CsInstallResourceName(&effectors, "Left Hand");
    CsInstallResourceName(&effectors, "Right Hand");
    return(effectors);

}

/******************************************************************************/

void ColavitaReinitialiseObjects(GENERIC_LIST *object_list)
{
    CsObject *object;

    object = CsGetObject(object_list, "auditory");
    CsResetObjectFeatures(object);
    CsInstallObjectFeatureFunctor(object, "tone off");

    object = CsGetObject(object_list, "visual");
    CsResetObjectFeatures(object);
    CsInstallObjectFeatureFunctor(object, "light off");

    object = CsGetObject(object_list, "l_button");
    CsResetObjectFeatures(object);
    CsInstallObjectFeatureFunctor(object, "off");

    object = CsGetObject(object_list, "r_button");
    CsResetObjectFeatures(object);
    CsInstallObjectFeatureFunctor(object, "off");

    if (MATHS == NORMALISED) {
        CsNormaliseObjects(object_list, 0);
    }
}

/******************************************************************************/

void ColavitaReinitialiseResources(CsResource *effector_list)
{
    CsResource *effector;

    effector = CsGetResource(effector_list, "Left Hand");
    CsResetResourceFeatures(effector);

    effector = CsGetResource(effector_list, "Right Hand");
    CsResetResourceFeatures(effector);

    if (MATHS == NORMALISED) {
        CsNormaliseResources(effector_list, 0);
    }
}

/******************************************************************************/

