/*******************************************************************************

  This file contains definitions and descriptions of the schema hierarchy and
  definitions of the triggers between schemas, objects, and effectors.

*******************************************************************************/

#include "cs.h"

extern CsSchema *CsInstallSchemaNode(CsSchema  **schemas, char *schema_name, char *schema_goal, int level, ActionType action);
extern void CsInstallSchemaTriggeringFunctions(CsSchema *schema, BOOLEAN (*obj)(), BOOLEAN (*res)(), double (*sit)());
extern void CsInstallSchemaSubgoal(CsSchema *schema, char *subgoal, ArgumentFunction arg_function, GoalType sort);
extern void CsInstallOrderingConstraint(CsSchema *schema, char *subgoal1, char *subgoal2);
extern void CsInstallSchemaResource(CsSchema *schema, Resource resource);

/******************************************************************************/

extern BOOLEAN RtNoStimulusTriggersObject(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN RtAuditoryStimulusTriggersObject(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN RtVisualStimulusTriggersObject(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN RtPrepareTriggersObject(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN RtStimulusPresentTriggersObject(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN RtPressLeftTriggersObject(CsObject *object, ArgumentFunction function, int cycle);
extern BOOLEAN RtPressRightTriggersObject(CsObject *object, ArgumentFunction function, int cycle);

/******************************************************************************/

extern BOOLEAN RtNoStimulusTriggersResource(CsResource *effector, int cycle);
extern BOOLEAN RtAuditoryStimulusTriggersResource(CsResource *effector, int cycle);
extern BOOLEAN RtVisualStimulusTriggersResource(CsResource *effector, int cycle);
extern BOOLEAN RtPrepareTriggersResource(CsResource *effector, int cycle);
extern BOOLEAN RtStimulusPresentTriggersResource(CsResource *effector, int cycle);
extern BOOLEAN RtPressLeftTriggersResource(CsResource *effector, int cycle);
extern BOOLEAN RtPressRightTriggersResource(CsResource *effector, int cycle);

/******************************************************************************/

extern double RtSituationTriggersNoStimulus(int cycle);
extern double RtSituationTriggersStimulusPresentA(int cycle);
extern double RtSituationTriggersStimulusPresentV(int cycle);
extern double RtSituationTriggersStimulusPresent(int cycle);
extern double RtSituationTriggersPrepare(int cycle);
extern double RtSituationTriggersAuditoryStimulus(int cycle);
extern double RtSituationTriggersVisualStimulus(int cycle);
extern double RtSituationTriggersPressLeft(int cycle);
extern double RtSituationTriggersPressRight(int cycle);

/******************************************************************************/

CsSchema *CsInitialiseSimpleAuditoryRTSchemas()
{
    CsSchema *schemas = NULL;
    CsSchema *schema;

/********* Level 3 ************************************************************/

    schema = CsInstallSchemaNode(&schemas, "No Stimulus", "Decide", 3, NONE);
    CsInstallSchemaSubgoal(schema, "Prepare", PRESS, CRUX);
    CsInstallSchemaTriggeringFunctions(schema, RtNoStimulusTriggersObject, RtNoStimulusTriggersResource, RtSituationTriggersNoStimulus);

    schema = CsInstallSchemaNode(&schemas, "Auditory Stimulus", "Decide", 3, NONE);
    CsInstallSchemaSubgoal(schema, "Press Button", PRESS, CRUX);
    CsInstallSchemaTriggeringFunctions(schema, RtAuditoryStimulusTriggersObject, RtAuditoryStimulusTriggersResource, RtSituationTriggersStimulusPresentA);

/********* Level 1 ************************************************************/

    schema = CsInstallSchemaNode(&schemas, "Wait for it", "Prepare", 1, PREPARE);
    CsInstallSchemaResource(schema,      L_HAND);
    CsInstallSchemaResource(schema,      R_HAND);
    CsInstallSchemaTriggeringFunctions(schema, RtPrepareTriggersObject, RtPrepareTriggersResource, RtSituationTriggersPrepare);

    schema = CsInstallSchemaNode(&schemas, "Push Left", "Press Button", 1, PUSH);
    CsInstallSchemaResource(schema,      L_HAND);
    CsInstallSchemaTriggeringFunctions(schema, RtPressLeftTriggersObject, RtPressLeftTriggersResource, RtSituationTriggersPressLeft);

/******************************************************************************/

    return(schemas);
}

CsSchema *CsInitialiseSimpleVisualRTSchemas()
{
    CsSchema *schemas = NULL;
    CsSchema *schema;

/********* Level 3 ************************************************************/

    schema = CsInstallSchemaNode(&schemas, "No Stimulus", "Decide", 3, NONE);
    CsInstallSchemaSubgoal(schema, "Prepare", PRESS, CRUX);
    CsInstallSchemaTriggeringFunctions(schema, RtNoStimulusTriggersObject, RtNoStimulusTriggersResource, RtSituationTriggersNoStimulus);

    schema = CsInstallSchemaNode(&schemas, "Visual Stimulus", "Decide", 3, NONE);
    CsInstallSchemaSubgoal(schema, "Press Button", PRESS, CRUX);
    CsInstallSchemaTriggeringFunctions(schema, RtVisualStimulusTriggersObject, RtVisualStimulusTriggersResource, RtSituationTriggersStimulusPresentV);

/********* Level 1 ************************************************************/

    schema = CsInstallSchemaNode(&schemas, "Wait for it", "Prepare", 1, PREPARE);
    CsInstallSchemaResource(schema,      L_HAND);
    CsInstallSchemaResource(schema,      R_HAND);
    CsInstallSchemaTriggeringFunctions(schema, RtPrepareTriggersObject, RtPrepareTriggersResource, RtSituationTriggersPrepare);

    schema = CsInstallSchemaNode(&schemas, "Push Right", "Press Button", 1, PUSH);
    CsInstallSchemaResource(schema,      R_HAND);
    CsInstallSchemaTriggeringFunctions(schema, RtPressRightTriggersObject, RtPressRightTriggersResource, RtSituationTriggersPressRight);

/******************************************************************************/

    return(schemas);
}

CsSchema *CsInitialiseChoiceRTSchemas()
{
    CsSchema *schemas = NULL;
    CsSchema *schema;

/********* Level 3 ************************************************************/

    schema = CsInstallSchemaNode(&schemas, "No Stimulus", "Decide", 3, NONE);
    CsInstallSchemaSubgoal(schema, "Prepare", PRESS, CRUX);
    CsInstallSchemaTriggeringFunctions(schema, RtNoStimulusTriggersObject, RtNoStimulusTriggersResource, RtSituationTriggersNoStimulus);

    schema = CsInstallSchemaNode(&schemas, "Stimulus Present", "Decide", 3, NONE);
    CsInstallSchemaSubgoal(schema, "Choose Button", PRESS, CRUX);
    CsInstallSchemaTriggeringFunctions(schema, RtStimulusPresentTriggersObject, RtStimulusPresentTriggersResource, RtSituationTriggersStimulusPresent);

/********* Level 2 ************************************************************/

    schema = CsInstallSchemaNode(&schemas, "Auditory Stimulus", "Choose Button", 2, NONE);
    CsInstallSchemaSubgoal(schema, "Press Left Button", PRESS, CRUX);
    CsInstallSchemaTriggeringFunctions(schema, RtAuditoryStimulusTriggersObject, RtAuditoryStimulusTriggersResource, RtSituationTriggersAuditoryStimulus);

    schema = CsInstallSchemaNode(&schemas, "Visual Stimulus", "Choose Button", 2, NONE);
    CsInstallSchemaSubgoal(schema, "Press Right Button", PRESS, CRUX);
    CsInstallSchemaTriggeringFunctions(schema, RtVisualStimulusTriggersObject, RtVisualStimulusTriggersResource, RtSituationTriggersVisualStimulus);

/********* Level 1 ************************************************************/

    schema = CsInstallSchemaNode(&schemas, "Wait for it", "Prepare", 1, PREPARE);
    CsInstallSchemaResource(schema,      L_HAND);
    CsInstallSchemaResource(schema,      R_HAND);
    CsInstallSchemaTriggeringFunctions(schema, RtPrepareTriggersObject, RtPrepareTriggersResource, RtSituationTriggersPrepare);

    schema = CsInstallSchemaNode(&schemas, "Push Left", "Press Left Button", 1, PUSH);
    CsInstallSchemaResource(schema,      L_HAND);
    CsInstallSchemaTriggeringFunctions(schema, RtPressLeftTriggersObject, RtPressLeftTriggersResource, RtSituationTriggersPressLeft);

    schema = CsInstallSchemaNode(&schemas, "Push Right", "Press Right Button", 1, PUSH);
    CsInstallSchemaResource(schema,      R_HAND);
    CsInstallSchemaTriggeringFunctions(schema, RtPressRightTriggersObject, RtPressRightTriggersResource, RtSituationTriggersPressRight);

/******************************************************************************/

    return(schemas);
}

/******************************************************************************/
