/******************************************************************************/

#include "cs.h"

/******************************************************************************/

extern double max_salience();
extern double net_influence();

/******************************************************************************/

extern BOOLEAN ColavitaIsVisualObject();
extern BOOLEAN ColavitaIsAuditoryObject();
extern BOOLEAN ColavitaIsVisualStimulus();
extern BOOLEAN ColavitaIsAuditoryStimulus();
extern BOOLEAN ColavitaIsVisualTarget();
extern BOOLEAN ColavitaIsAuditoryTarget();
extern BOOLEAN ColavitaIsLeftHand();
extern BOOLEAN ColavitaIsRightHand();
extern BOOLEAN ColavitaIsLeftButton();
extern BOOLEAN ColavitaIsRightButton();
extern BOOLEAN ColavitaIsPressedButton();
extern BOOLEAN ColavitaIsUnpressedButton();

extern BOOLEAN RtAwaitingResponse();

/******************************************************************************/

double RtSituationTriggersNoStimulus(int cycle)
{
    CsObject *stimulus;
    for (stimulus = objects; stimulus != NULL; stimulus = stimulus->next) {
        if ((ColavitaIsAuditoryStimulus(stimulus)) || (ColavitaIsVisualStimulus(stimulus))) {
            return(-1.0);
        }
    }
    return(1.0);
}

double RtSituationTriggersStimulusPresentA(int cycle)
{
    double total = 0.0;

    if (RtAwaitingResponse("l_button")) {
        CsObject *stimulus;
        for (stimulus = objects; stimulus != NULL; stimulus = stimulus->next) {
            if (ColavitaIsAuditoryStimulus(stimulus)) {
                total += net_influence(1, stimulus->activation[ATTEND][cycle-1]);
            }
        }
    }
    return(total * AUDITORY_ROUTE_STRENGTH);
}

double RtSituationTriggersStimulusPresentV(int cycle)
{
    double total = 0.0;

    if (RtAwaitingResponse("r_button")) {
        CsObject *stimulus;
        for (stimulus = objects; stimulus != NULL; stimulus = stimulus->next) {
            if (ColavitaIsVisualStimulus(stimulus)) {
                total += net_influence(1, stimulus->activation[ATTEND][cycle-1]);
            }
        }
    }
    return(total * VISUAL_ROUTE_STRENGTH);
}

double RtSituationTriggersStimulusPresent(int cycle)
{
    return(RtSituationTriggersStimulusPresentA(cycle) + RtSituationTriggersStimulusPresentV(cycle));
}

double RtSituationTriggersPrepare(int cycle)
{
    return(RtSituationTriggersNoStimulus(cycle));
}

double RtSituationTriggersAuditoryStimulus(int cycle)
{
    return(RtSituationTriggersStimulusPresentA(cycle));
}

double RtSituationTriggersVisualStimulus(int cycle)
{
    return(RtSituationTriggersStimulusPresentV(cycle));
}

double RtSituationTriggersPressLeft(int cycle)
{
    return(0.0);
    return(RtSituationTriggersStimulusPresentA(cycle));
}

double RtSituationTriggersPressRight(int cycle)
{
    return(0.0);
    return(RtSituationTriggersStimulusPresentV(cycle));
}

/******************************************************************************/

double RtSituationTriggersAuditoryObject(int cycle, ArgumentFunction f)
{
    if ((f == ATTEND) && (RtAwaitingResponse("l_button"))) {
        CsObject *stimulus;
        for (stimulus = objects; stimulus != NULL; stimulus = stimulus->next) {
            if (ColavitaIsAuditoryStimulus(stimulus)) {
                return(AUDITORY_ROUTE_STRENGTH);
            }
        }
    }
    return(0.0);
}

double RtSituationTriggersVisualObject(int cycle, ArgumentFunction f)
{
    if ((f == ATTEND) && (RtAwaitingResponse("r_button"))) {
        CsObject *stimulus;
        for (stimulus = objects; stimulus != NULL; stimulus = stimulus->next) {
            if (ColavitaIsVisualStimulus(stimulus)) {
                return(VISUAL_ROUTE_STRENGTH);
            }
        }
    }
    return(0.0);
}

double RtSituationTriggersButton(int cycle, ArgumentFunction f)
{
    return(0.0);
}

/******************************************************************************/

BOOLEAN RtNoStimulusTriggersObject(CsObject *object, ArgumentFunction function, int cycle)
{
    return(FALSE);
}

BOOLEAN RtAuditoryStimulusTriggersObject(CsObject *object, ArgumentFunction function, int cycle)
{
    if (function == ATTEND) {
        return(RtAwaitingResponse("l_button") && ColavitaIsAuditoryStimulus(object));
    }
    else if (function == PRESS) {
        return(ColavitaIsLeftButton(object) && !ColavitaIsUnpressedButton(object) && !ColavitaIsPressedButton(object));
    }
    else {
        return(FALSE);
    }
}

BOOLEAN RtVisualStimulusTriggersObject(CsObject *object, ArgumentFunction function, int cycle)
{
    if (function == ATTEND) {
        return(RtAwaitingResponse("r_button") && ColavitaIsVisualStimulus(object));
    }
    else if (function == PRESS) {
        return(ColavitaIsRightButton(object) && !ColavitaIsUnpressedButton(object) && !ColavitaIsPressedButton(object));
    }
    else {
        return(FALSE);
    }
}

BOOLEAN RtPrepareTriggersObject(CsObject *object, ArgumentFunction function, int cycle)
{
    if (function == ATTEND) {
        return(ColavitaIsAuditoryObject(object) || ColavitaIsVisualObject(object));
    }
    else if (function == PRESS) {
        return(ColavitaIsLeftButton(object) || ColavitaIsRightButton(object));
    }
    else {
        return(FALSE);
    }
}

BOOLEAN RtStimulusPresentTriggersObject(CsObject *object, ArgumentFunction function, int cycle)
{
    return(RtAuditoryStimulusTriggersObject(object, function, cycle) || RtVisualStimulusTriggersObject(object, function, cycle));
}

BOOLEAN RtPressLeftTriggersObject(CsObject *object, ArgumentFunction function, int cycle)
{
    if (function == ATTEND) {
        return(RtAwaitingResponse("l_button") && ColavitaIsAuditoryTarget(object));
    }
    else if (function == PRESS) {
        return(ColavitaIsLeftButton(object) && !ColavitaIsPressedButton(object));
    }
    else {
        return(FALSE);
    }
}

BOOLEAN RtPressRightTriggersObject(CsObject *object, ArgumentFunction function, int cycle)
{
    if (function == ATTEND) {
        return(RtAwaitingResponse("r_button") && ColavitaIsVisualTarget(object));
    }
    else if (function == PRESS) {
        return(ColavitaIsRightButton(object) && !ColavitaIsPressedButton(object));
    }
    else {
        return(FALSE);
    }
}

/******************************************************************************/

BOOLEAN RtNoStimulusTriggersResource(CsResource *effector, int cycle)
{
    return(FALSE);
}

BOOLEAN RtAuditoryStimulusTriggersResource(CsResource *effector, int cycle)
{
    return(RtAwaitingResponse("l_button") && ColavitaIsLeftHand(effector));
}

BOOLEAN RtVisualStimulusTriggersResource(CsResource *effector, int cycle)
{
    return(RtAwaitingResponse("r_button") && ColavitaIsRightHand(effector));
}

BOOLEAN RtPrepareTriggersResource(CsResource *effector, int cycle)
{
    return(TRUE);
}

BOOLEAN RtStimulusPresentTriggersResource(CsResource *effector, int cycle)
{
    return(RtAuditoryStimulusTriggersResource(effector, cycle) || RtVisualStimulusTriggersResource(effector, cycle));
}

BOOLEAN RtPressLeftTriggersResource(CsResource *effector, int cycle)
{
    return(RtAwaitingResponse("l_button") && ColavitaIsLeftHand(effector));
}

BOOLEAN RtPressRightTriggersResource(CsResource *effector, int cycle)
{
    return(RtAwaitingResponse("r_button") && ColavitaIsRightHand(effector));
}

/******** END OF FILE *********************************************************/


