/********* Types for Contention Scheduling in C *******************************/

/*******************************************************************************

NOTE
====

There are three interfaces to the basic cs program:
    cs is the basic cs system. It runs one trial with no frills.
    xcs is a cs with an X interface, using the Motif plotting widgets to
        graph activation profiles.
    mcs is cs with for multiple trials. It reruns cs a specified number of
        times, saving the output for later analysis.

    to make xvcs, do "make xvcs"
    to make mcs, do "make mcs"
    to make cs, do "make cs"

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MaxCycles 2500
#define MaxFunctions 5

typedef enum action_type {NONE, PICK_UP, PUT_DOWN, DIP_SPOON, EMPTY_SPOON, EAT_FROM_SPOON,
                          STIR, POUR, TEAR, SCREW_OPEN, SCREW_CLOSED,
			  SWAP, DRINK, PREPARE, PUSH, RELAX} ActionType;
typedef enum resource {HAND = 1, L_HAND, R_HAND} Resource;
typedef enum argument_function {EMPTY = -1, IMPLEMENT, SOURCE, TARGET, PRESS, ATTEND, ANY} ArgumentFunction;

/******************************************************************************/

char *function_name[MaxFunctions];

typedef enum boolean {FALSE, TRUE} BOOLEAN;
typedef enum maths_type {TANH, SIGMOID, HALFX, M_R, HOUGHTON, NORMALISED} MathsType;
typedef enum task_type {COFFEE_PREPARATION, CEREAL_CONSUMPTION, BREAKFAST, SIMPLE_VISUAL_RT, SIMPLE_AUDITORY_RT, CHOICE_RT, CONFLICT_RT} TaskType;
typedef enum message_type {ACTION, SELECTION, WORLD, WARNING} MessageType;
typedef enum term_type {GROUND, VARIABLE, LIST} TermType;
typedef enum goal_type {PRECONDITION, CRUX, CLEANUP} GoalType;

/* Parameters: */

extern double SCHEMA_SI, SCHEMA_PE, SCHEMA_SL,
              OBJECT_SI, OBJECT_PE, OBJECT_SL, 
              EFFECT_SI, EFFECT_PE, EFFECT_SL, 
              MIN_ACT, MAX_ACT, REST_ACT, PERSISTENCE, SELECTION_THRESHOLD, NOISE_SD;

extern int VISUAL_LAG, AUDITORY_LAG;
extern double MS_PER_CYCLE, VISUAL_ROUTE_STRENGTH, AUDITORY_ROUTE_STRENGTH;
extern MathsType MATHS;

/* Statistics: */

extern int ARG_SEL_FAIL_COUNT, ACTION_FAIL_COUNT;

typedef struct GenericList {
    void *head;
    struct GenericList *next;
} GENERIC_LIST;

typedef struct cs_schema {
    double activation[MaxCycles];
    int level;
    char *name;
    char *goal;
    BOOLEAN selected;
    BOOLEAN executed;
    ActionType corresponding_action;
    GENERIC_LIST *res;
    struct subgoal *children;
    BOOLEAN (*object_triggering_function)();
    BOOLEAN (*effector_triggering_function)();
    double (*triggering_situation)();
    ArgumentFunction default_argument_function;
    FILE *output_stream;    /* Only used when saving activations */
    BOOLEAN highlighted;    /* Only used in X_VERSION */
    struct cs_schema *next;
} CsSchema;

typedef struct cs_object {
    char *name;
    double activation[MaxFunctions][MaxCycles];
    GENERIC_LIST *features;
    double (*triggering_situation)();
    FILE *output_stream;    /* Only used when saving activations */
    BOOLEAN highlighted;    /* Only used in X_VERSION */
    struct cs_object *next;
} CsObject;

typedef struct cs_resource {
    char *name;
    double activation[MaxCycles];
    GENERIC_LIST *features;
    FILE *output_stream;    /* Only used when saving activations */
    BOOLEAN highlighted;    /* Only used in X_VERSION */
    struct cs_resource *next;
} CsResource;

typedef struct subgoal {
    char *subgoal;
    ArgumentFunction arg_function;
    BOOLEAN achieved;
    GoalType sort;
    GENERIC_LIST *prior_subgoals;
    struct subgoal *next;
} Subgoal;

typedef struct Term {
    TermType sort;
    char *functor;
    GENERIC_LIST *arguments;
} TERM;

typedef struct OutputFlags {
    int action;
    int selection;
    int world;
    int warning;
    int activations;
} OUTPUTFLAGS;

extern OUTPUTFLAGS display_message;

extern CsSchema   *schemas;
extern CsObject   *objects;
extern CsResource *resources;

#define MAX_LEVEL   4

extern char *copy_string(char *string);

/******************************************************************************/

