#include "cs.h"

extern void write_term_list();

/* The directory in which activation files are written (if data is being logged): */
#define OUTPUT_DIR "/tmp"

/******************************************************************************/

int ARG_SEL_FAIL_COUNT, ACTION_FAIL_COUNT;

/******************************************************************************/

void CsInitialiseStats()
{
    ARG_SEL_FAIL_COUNT = 0;
    ACTION_FAIL_COUNT = 0;
}

/******************************************************************************/
/******** DATA LOGGING ********************************************************/

BOOLEAN CsOpenActivationFiles()
{
    CsSchema   *tmp_s;
    CsObject   *tmp_o;
    CsResource *tmp_r;
    char filename[100];

    for (tmp_s = schemas; tmp_s != NULL; tmp_s = tmp_s->next) {
        sprintf(filename, "%s/schema_activation.%s", OUTPUT_DIR, tmp_s->name);
        if ((tmp_s->output_stream = fopen(filename, "w")) == NULL) {
            fprintf(stderr, "WARNING: Cannot open save file (%s)\n", filename);
            fprintf(stderr, "Data will not be saved\n");
            return(FALSE);
        }
    }
    for (tmp_o = objects; tmp_o != NULL; tmp_o = tmp_o->next) {
        sprintf(filename, "%s/object_salience.%s", OUTPUT_DIR, tmp_o->name);
        if ((tmp_o->output_stream = fopen(filename, "w")) == NULL) {
            fprintf(stderr, "WARNING: Cannot open save file (%s)\n", filename);
            fprintf(stderr, "Data will not be saved\n");
            return(FALSE);
        }
    }
    for (tmp_r = resources; tmp_r != NULL; tmp_r = tmp_r->next) {
        sprintf(filename, "%s/effector_activation.%s", OUTPUT_DIR, tmp_r->name);
        if ((tmp_r->output_stream = fopen(filename, "w")) == NULL) {
            fprintf(stderr, "WARNING: Cannot open save file (%s)\n", filename);
            fprintf(stderr, "Data will not be saved\n");
            return(FALSE);
        }
    }
    return(TRUE);
}    

void CsSaveActivations(int cycle)
{
    CsSchema   *tmp_s;
    CsObject   *tmp_o;
    CsResource *tmp_r;
    int f = SOURCE;

    for (tmp_s = schemas; tmp_s != NULL; tmp_s = tmp_s->next) {
        fprintf(tmp_s->output_stream, "%3d %6.4f\n", cycle, tmp_s->activation[cycle]);
    }
    for (tmp_o = objects; tmp_o != NULL; tmp_o = tmp_o->next) {
        fprintf(tmp_o->output_stream, "%3d %6.4f ", cycle, tmp_o->activation[f][cycle]);
        fprintf(tmp_o->output_stream, "\n");
    }
    for (tmp_r = resources; tmp_r != NULL; tmp_r = tmp_r->next) {
        fprintf(tmp_r->output_stream, "%3d %6.4f\n", cycle, tmp_r->activation[cycle]);
    }
}

void CsCloseActivationFiles()
{
    CsSchema   *tmp_s;
    CsObject   *tmp_o;
    CsResource *tmp_r;

    for (tmp_s = schemas; tmp_s != NULL; tmp_s = tmp_s->next) {
        fclose(tmp_s->output_stream);
    }
    for (tmp_o = objects; tmp_o != NULL; tmp_o = tmp_o->next) {
        fclose(tmp_o->output_stream);
    }
    for (tmp_r = resources; tmp_r != NULL; tmp_r = tmp_r->next) {
        fclose(tmp_r->output_stream);
    }
}    

/******************************************************************************/
/******** MESSAGES ************************************************************/

void CsMessage(MessageType category, int cycle, char *buffer)
{
    switch (category) {
    case ACTION:
        if (display_message.action) {
            fprintf(stdout, "\n%-5d%s", cycle, buffer);
            fflush(stdout);
        }
        break;
    case SELECTION:
        if (display_message.selection) {
            fprintf(stdout, "\n%-5d%s", cycle, buffer);
            fflush(stdout);
        }
        break;
    case WORLD:
        if (display_message.world) {
            fprintf(stdout, "\n%-5d%s", cycle, buffer);
            fflush(stdout);
        }
        break;
    case WARNING:
        if (display_message.warning) {
            fprintf(stdout, "\n%-5d%s", cycle, buffer);
            fflush(stdout);
        }
        break;
    }
}

/*----------------------------------------------------------------------------*/

void CsSelectionMessage(int cycle, CsSchema *schema, BOOLEAN flag)
{
    int tab;
    char buffer[80];

    tab = 2 * (5 - schema->level);
    sprintf(buffer, "%*s%c%s", tab, "", (flag?'+':'-'), schema->name);
    CsMessage(SELECTION, cycle, buffer);
}

/******************************************************************************/

void CsArgumentFailureMessage(int cycle, char *action)
{
    char buffer[80];

    sprintf(buffer, "        Failed to select arguments for %s.", action);
    CsMessage(WARNING, cycle, buffer);
    ARG_SEL_FAIL_COUNT++;
}

void CsActionFailureMessage(int cycle, char *reason)
{
    char buffer[80];

    sprintf(buffer, "        (%s!)", reason);
    CsMessage(WARNING, cycle, buffer);
    ACTION_FAIL_COUNT++;
}

/******************************************************************************/

void EnvPreActionTrace(int cycle, char *name, GENERIC_LIST *features)
{
    char buf[80];

    if (display_message.world) {
        sprintf(buf, "      %s: ", name);
        CsMessage(WORLD, cycle, buf);
        fprintf(stdout, "[");
        write_term_list(stdout, features);
        fprintf(stdout, "]");
        fflush(stdout);
    }
}

void EnvPostActionTrace(GENERIC_LIST *features)
{
    if (display_message.world) {
        fprintf(stdout, "          -->  [");
        write_term_list(stdout, features);
        fprintf(stdout, "]");
        fflush(stdout);
    }
}

/******************************************************************************/
/******** TABULAR OUTPUT ******************************************************/

static char *get_subgoal_string(CsSchema *s)
{
    char *str;
    int j = 0;
    Subgoal *tmp;

    if ((str = (char *)malloc(20 * sizeof(char))) == NULL) {
        fprintf(stderr, "WARNING: Insufficient memory for malloc (get_subgoal_string)");
    }
    else {
        str[j++] = '[';
        for (tmp = s->children; tmp != NULL; tmp = tmp->next) {
            if (j < 19) {
                if (tmp->achieved) {
                    str[j++] = '+';
                }
                else {
                    str[j++] = '-';
                }
            }
            if (j < 19) {
                str[j++] = ',';
            }
        }
        str[((j > 18) ? 18 : --j)] = ']';
        str[++j] = '\0';
    }
    return(str);
}

/******************************************************************************/

static char *get_selector(CsSchema *i)
{
    extern CsSchema *selected_parent_of();
    CsSchema *p;

    if (i->selected) {
        if ((p = selected_parent_of(i)) != NULL) {
            return(p->name);
        }
        else {
            return("intention");
        }
    }
    else {
        return("");
    }
}

/******************************************************************************/

static void list_schema_by_levels(int cycle, FILE *file)
{
    extern double SchemaSelfInfluence();
    extern double SchemaLateralInfluence();
    extern double SchemaIntentionalInfluence();
    extern double SchemaEnvironmentalInfluence();

    int level;
    CsSchema *i;

    for (level=MAX_LEVEL; (level > 0); level--) {
        for (i = schemas; i != NULL; i = i->next) {
            if (i->level == level) {
                if (i->selected) {
                    fprintf(file, "* %-20s%-18s%-10s", i->name, get_selector(i), get_subgoal_string(i));
                }
                else {
                    fprintf(file, "  %-20s%28s", i->name, "");
		}
                fprintf(file, "(%6.3f %6.3f %6.3f %6.3f)     %8.5f\n", SchemaSelfInfluence(cycle+1, i), SchemaLateralInfluence(cycle+1, i), SchemaIntentionalInfluence(cycle+1, i), SchemaEnvironmentalInfluence(cycle+1, i), i->activation[cycle]);
            }
        }
        fprintf(file, "................................................................................\n");        
    }
}

static void list_function_activations(int cycle, CsObject *obj, FILE *file)
{
    extern double ObjectSelfInfluence();
    extern double ObjectLateralInfluence();
    extern double ObjectIntentionalInfluence();

    int f;

    for (f = 0; f < MaxFunctions; f++) {
        if (function_name[f] != NULL) {
            fprintf(file, "  %10s%38s(%6.3f %6.3f %6.3f)            %8.5f\n", function_name[f], "",
                ObjectSelfInfluence(cycle+1, obj, f),
    	        ObjectLateralInfluence(cycle+1, obj, f),
                ObjectIntentionalInfluence(cycle+1, obj, f), 
                obj->activation[f][cycle]);
        }
    }
}

static BOOLEAN trace_attribute(TERM *term)
{
    if (term != NULL) {
        return( (strcmp(term->functor, "open") == 0) ||
                (strcmp(term->functor, "closed") == 0) ||
                (strcmp(term->functor, "holds") == 0) ||
                (strcmp(term->functor, "pressed") == 0) ||
                (strcmp(term->functor, "unpressed") == 0) ||
                (strcmp(term->functor, "off") == 0) ||
                (strcmp(term->functor, "high") == 0) ||
                (strcmp(term->functor, "low") == 0) ||
                (strcmp(term->functor, "red") == 0) ||
                (strcmp(term->functor, "white") == 0) ||
                (strcmp(term->functor, "blue") == 0) ||
                (strcmp(term->functor, "green") == 0) );
    }
    else {
        return(FALSE);
    }
}

static void list_trace_attributes(GENERIC_LIST *feature_ptr, FILE *file)
{
    while (feature_ptr != NULL) {
        if (trace_attribute(((TERM *)(feature_ptr->head)))) {

            GENERIC_LIST *tmp = feature_ptr->next;

            fprintf(file, " ");

            feature_ptr->next = NULL;
            write_term_list(file, feature_ptr);
            feature_ptr->next = tmp;
        }
        feature_ptr = feature_ptr->next;
    }
}

static void list_object_activations(int cycle, FILE *file)
{
    CsObject *tmp;
    char buf[20];

    for (tmp = objects; tmp != NULL; tmp = tmp->next) {
        sprintf(buf, " %s:", tmp->name);
        fprintf(file, "%-20s", buf);
        list_trace_attributes(tmp->features, file);
        fprintf(file, "\n");
        list_function_activations(cycle, tmp, file);
    }
}

static void list_effector_activations(int cycle, FILE *file)
{
    extern double ResourceIntentionalInfluence();
    CsResource *tmp;
    char buf[20];

    for (tmp=resources; tmp != NULL; tmp = tmp->next) {
        sprintf(buf, " %s:", tmp->name);
        fprintf(file, "%-19s%8.5f ", buf, tmp->activation[cycle]);
        fprintf(file, "                                          %8.5f     ", ResourceIntentionalInfluence(cycle+1, tmp));
        list_trace_attributes(tmp->features, file);
        fprintf(file, "\n");
    }
}

static int width(int n)
{
    return((n >= 10) ? (1 + width((int) n/10)) : 1);
}

void CsPostCycleTrace(int cycle, FILE *file)
{
    char pad[5] = "     ";
    int w = width(cycle);

    pad[4 - w] = '\0';

    fprintf(file, "%d ***********************************************%sSelf   Lat    Par    Env\n", cycle, pad);
    list_schema_by_levels(cycle, file);
    list_object_activations(cycle, file);
    fprintf(file, "................................................................................\n");
    list_effector_activations(cycle, file);
    fprintf(file, "................................................................................\n");
    fprintf(file, "\n");
}

/******** END OF FILE *********************************************************/
