/******************************************************************************/

#include "cs.h"

/******************************************************************************/

TERM *get_feature(GENERIC_LIST *features, char *functor)
{
    while ((features != NULL) && (strcmp(((TERM *)(features->head))->functor, functor) != 0)) {
        features = features->next;
    }

    if (features == NULL) {
        return(NULL);
    }
    else {
        return(((TERM *)(features->head)));
    }
}

BOOLEAN has_feature(GENERIC_LIST *features, char *functor)
{
    return(get_feature(features, functor) != NULL);
}

/******************************************************************************/


double max(double i, double j)
{
    return((i > j) ? i : j);
}

double max_salience(CsObject *object, int cycle)
{
    return(max(max(object->activation[IMPLEMENT][cycle], object->activation[SOURCE][cycle]), object->activation[TARGET][cycle]));
}

CsObject *most_salient_object(int cycle, CsObject *previous, CsObject *current, ArgumentFunction function)
{
    if (previous == NULL) {
        return(current);
    }
    else if (function != ANY) {
        if (previous->activation[function][cycle] < current->activation[function][cycle]) {
            return(current);
        }
        else {
            return(previous);
        }
    }
    else if (max_salience(previous, cycle) < max_salience(current, cycle)) {
        return(current);
    }
    else {
        return(previous);
    }
}

CsResource *most_salient_effector(int cycle, CsResource *previous, CsResource *current)
{
    if ((previous == NULL) || (previous->activation[cycle] < current->activation[cycle])) {
        return(current);
    }
    else {
        return(previous);
    }
}

/******************************************************************************/

ArgumentFunction get_argument_function(CsSchema *schema)
{
    extern CsSchema *selected_parent_of();
    CsSchema *parent;
    Subgoal *tmp;

    if ((parent = selected_parent_of(schema)) == NULL) {
        return(schema->default_argument_function);
    }
    else {
        for (tmp = parent->children; tmp != NULL; tmp = tmp->next) {
            if (strcmp(tmp->subgoal, schema->goal) == 0) {
                return(tmp->arg_function);
            }
        }
        return(ANY);
    }
}

/******************************************************************************/

BOOLEAN object_can_serve_function(CsObject *object, ArgumentFunction function)
{
    extern BOOLEAN CoffeeIsContainer(CsObject *object);
    extern BOOLEAN CoffeeIsSource(CsObject *object);
    extern BOOLEAN CoffeeIsSpoon(CsObject *object);
    extern BOOLEAN CoffeeIsStirrer(CsObject *object);

    extern BOOLEAN ColavitaIsVisualStimulus(CsObject *object);
    extern BOOLEAN ColavitaIsAuditoryStimulus(CsObject *object);
    extern BOOLEAN ColavitaIsLeftButton(CsObject *object);
    extern BOOLEAN ColavitaIsRightButton(CsObject *object);

    switch (function) {
        case IMPLEMENT:
            return(CoffeeIsSpoon(object) || CoffeeIsStirrer(object));
        case SOURCE:
            return(CoffeeIsSource(object));
        case TARGET:
            return(CoffeeIsContainer(object));
        case ATTEND:
            return(ColavitaIsAuditoryStimulus(object) || ColavitaIsVisualStimulus(object));
        case PRESS:             
            return(ColavitaIsLeftButton(object) || ColavitaIsRightButton(object));
        default:
            return(FALSE);
    }
}

/******************************************************************************/

extern void pick_up(int cycle, CsObject *argument, CsResource *effector);
extern void put_down(int cycle, CsResource *effector);
extern void dip_spoon(int cycle, CsObject *argument, CsResource *effector);
extern void empty_spoon(int cycle, CsObject *argument, CsResource *effector);
extern void consume_spoon_contents(int cycle, CsResource *effector);
extern void stir(int cycle, CsObject *argument, CsResource *effector);
extern void pour(int cycle, CsObject *argument, CsResource *effector);
extern void tear(int cycle, CsResource *effector1, CsResource *effector2);
extern void screw_open(int cycle, CsResource *effector1, CsResource *effector2);
extern void screw_closed(int cycle, CsResource *effector1, CsResource *effector2);
extern void swap(int cycle, CsResource *effector1, CsResource *effector2);
extern void drink(int cycle, CsResource *effector);

extern BOOLEAN CoffeeIsOpen(CsObject *object);
extern BOOLEAN CoffeeIsHand(CsResource *effector);
extern BOOLEAN CoffeeIsEmptyHand(CsResource *effector);
extern BOOLEAN CoffeeIsHeld(CsObject *object);
extern BOOLEAN CoffeeIsSource(CsObject *object);
extern BOOLEAN CoffeeIsContainer(CsObject *object);
extern BOOLEAN CoffeeIsEmptyContainer(CsObject *source);
extern TERM *CoffeeContentsOfHand(CsResource *hand);
extern BOOLEAN CoffeeItemsIncludeSpoon(TERM *list);
extern BOOLEAN CoffeeItemsIncludeScrewableContainer(TERM *list);
extern BOOLEAN CoffeeItemsIncludeScrewableLid(TERM *list);
extern BOOLEAN CoffeeItemsIncludeOpenSource(TERM *list);
extern BOOLEAN CoffeeItemsIncludeUnscrewableContainer(TERM *list);
extern BOOLEAN CoffeeItemsIncludeStirrer(TERM *list);
extern BOOLEAN CoffeeItemsIncludeTearableContainer(TERM *list);

/*----------------------------------------------------------------------------*/

extern void press(int cycle, CsResource *effector);
extern void prepare(int cycle, CsResource *effector);
extern void relax(int cycle, CsResource *effector);

extern BOOLEAN ColavitaIsHand(CsResource *effector);

/*----------------------------------------------------------------------------*/

void execute_action(CsSchema *schema, int cycle)
{
    CsObject *argument, *argument1;
    CsResource *effector, *effector1, *effector2;
    ArgumentFunction function;

    switch(schema->corresponding_action) {
    case NONE:
        fprintf(stderr, "WARNING: Attempting to execute non-primitive action.\n");
        break;
    case PICK_UP:    /* pick_up(ObjectWithinReach, EmptyHand) */
        argument = NULL; effector = NULL;
        function = get_argument_function(schema);
        for (effector1 = resources; effector1 != NULL; effector1 = effector1->next) {
            if (CoffeeIsEmptyHand(effector1)) {
                effector = most_salient_effector(cycle, effector, effector1);
            }
        }
        for (argument1 = objects; argument1 != NULL; argument1 = argument1->next) {
            if (!CoffeeIsHeld(argument1)) {
                argument = most_salient_object(cycle, argument, argument1, function);
            }
        }
        pick_up(cycle, argument, effector);
        break;
    case PUT_DOWN:    /* put_down(HandContainingObject) */
        effector = NULL;
        for (effector1 = resources; effector1 != NULL; effector1 = effector1->next) {
            if (CoffeeIsHand(effector1) && !(CoffeeIsEmptyHand(effector1))) {
                effector = most_salient_effector(cycle, effector, effector1);
            }
        }
        put_down(cycle, effector);
        break;
    case DIP_SPOON:    /* dip_spoon(HandContainingSpoon, OpenSource) */
        argument = NULL; effector = NULL;
        function = get_argument_function(schema);
        for (argument1 = objects; argument1 != NULL; argument1 = argument1->next) {
            if (CoffeeIsOpen(argument1) && CoffeeIsSource(argument1)) {
                argument = most_salient_object(cycle, argument, argument1, function);
            }
        }
        for (effector1 = resources; effector1 != NULL; effector1 = effector1->next) {
            if (CoffeeIsHand(effector1) && CoffeeItemsIncludeSpoon(CoffeeContentsOfHand(effector1))) {
                effector = most_salient_effector(cycle, effector, effector1);
            }
        }
        dip_spoon(cycle, argument, effector);
        break;
    case EMPTY_SPOON:    /* empty_spoon(HandContainingSpoon, OpenIsEmptyContainer) */
        argument = NULL; effector = NULL;
        function = get_argument_function(schema);
        for (argument1 = objects; argument1 != NULL; argument1 = argument1->next) {
            if (CoffeeIsOpen(argument1) && CoffeeIsContainer(argument1)) {
                argument = most_salient_object(cycle, argument, argument1, function);
            }
        }
        for (effector1 = resources; effector1 != NULL; effector1 = effector1->next) {
            if (CoffeeIsHand(effector1) && CoffeeItemsIncludeSpoon(CoffeeContentsOfHand(effector1))) {
                effector = most_salient_effector(cycle, effector, effector1);
            }
        }
        empty_spoon(cycle, argument, effector);
        break;
    case EAT_FROM_SPOON:
        effector = NULL;
        for (effector1 = resources; effector1 != NULL; effector1 = effector1->next) {
            if (CoffeeIsHand(effector1) && CoffeeItemsIncludeSpoon(CoffeeContentsOfHand(effector1))) {
                effector = most_salient_effector(cycle, effector, effector1);
            }
        }
        consume_spoon_contents(cycle, effector);
        break;
    case STIR:    /* stir(HandContainingStirrer, NonEmptyOpenContainer) */
        argument = NULL; effector = NULL;
        function = get_argument_function(schema);
        for (argument1 = objects; argument1 != NULL; argument1 = argument1->next) {
            if (CoffeeIsOpen(argument1) && CoffeeIsContainer(argument1) && !(CoffeeIsEmptyContainer(argument1))) {
                argument = most_salient_object(cycle, argument, argument1, function);
            }
        }
        for (effector1 = resources; effector1 != NULL; effector1 = effector1->next) {
            if (CoffeeIsHand(effector1) && CoffeeItemsIncludeStirrer(CoffeeContentsOfHand(effector1))) {
                effector = most_salient_effector(cycle, effector, effector1);
            }
        }
        stir(cycle, argument, effector);
        break;
    case POUR:    /* pour(HandContainingOpenSource, OpenIsEmptyContainer) */
        argument = NULL; effector = NULL;
        function = get_argument_function(schema);
        for (argument1 = objects; argument1 != NULL; argument1 = argument1->next) {
            if (CoffeeIsOpen(argument1) && CoffeeIsContainer(argument1)) {
                argument = most_salient_object(cycle, argument, argument1, function);
            }
        }
        for (effector1 = resources; effector1 != NULL; effector1 = effector1->next) {
            if (CoffeeIsHand(effector1) && CoffeeItemsIncludeOpenSource(CoffeeContentsOfHand(effector1))) {
                effector = most_salient_effector(cycle, effector, effector1);
            }
        }
        pour(cycle, argument, effector);
        break;
    case TEAR:    /* tear(HandContainingPacket, EmptyHand) */
        effector1 = NULL; effector2 = NULL;
        for (effector = resources; effector != NULL; effector = effector->next) {
            if (CoffeeIsHand(effector) && CoffeeItemsIncludeTearableContainer(CoffeeContentsOfHand(effector))) {
                effector1 = most_salient_effector(cycle, effector1, effector);
            }
            if (CoffeeIsEmptyHand(effector)) {
                effector2 = most_salient_effector(cycle, effector2, effector);
            }
        }
        tear(cycle, effector1, effector2);
        break;
    case SCREW_OPEN:    /* screw_open(HandContainingJar, EmptyHand) */
        effector1 = NULL; effector2 = NULL;
        for (effector = resources; effector != NULL; effector = effector->next) {
            if (CoffeeIsHand(effector) && CoffeeItemsIncludeUnscrewableContainer(CoffeeContentsOfHand(effector))) {
                effector1 = most_salient_effector(cycle, effector1, effector);
            }
            if (CoffeeIsEmptyHand(effector)) {
                effector2 = most_salient_effector(cycle, effector2, effector);
            }
        }
        screw_open(cycle, effector1, effector2);
        break;
    case SCREW_CLOSED:    /* screw_closed(HandContainingJar, HandContainingScrewLid) */
        effector1 = NULL; effector2 = NULL;
        for (effector = resources; effector != NULL; effector = effector->next) {
            if (CoffeeIsHand(effector) && CoffeeItemsIncludeScrewableContainer(CoffeeContentsOfHand(effector))) {
                effector1 = most_salient_effector(cycle, effector1, effector);
            }
            if (CoffeeIsHand(effector) && CoffeeItemsIncludeScrewableLid(CoffeeContentsOfHand(effector))) {
                effector2 = most_salient_effector(cycle, effector2, effector);
            }
        }
        screw_closed(cycle, effector1, effector2);
        break;
    case SWAP:    /* swap(HandContainingObject, EmptyHand) */
        effector1 = NULL; effector2 = NULL;
        for (effector = resources; effector != NULL; effector = effector->next) {
            if (CoffeeIsEmptyHand(effector)) {
                effector1 = most_salient_effector(cycle, effector1, effector);
            }
            if (CoffeeIsHand(effector) && !(CoffeeIsEmptyHand(effector))) {
                effector2 = most_salient_effector(cycle, effector2, effector);
            }
        }
        swap(cycle, effector1, effector2);
        break;
    case DRINK:    /* drink(HandContainingOpenSource) */
        effector = NULL;
        for (effector1 = resources; effector1 != NULL; effector1 = effector1->next) {
            if (CoffeeIsHand(effector1) && CoffeeItemsIncludeOpenSource(CoffeeContentsOfHand(effector1))) {
                effector = most_salient_effector(cycle, effector, effector1);
            }
        }
        drink(cycle, effector);
        break;
    case PREPARE:    /* prepare(ActiveHand) */
        effector = NULL;
        for (effector1 = resources; effector1 != NULL; effector1 = effector1->next) {
            if (ColavitaIsHand(effector1)) {
                effector = most_salient_effector(cycle, effector, effector1);
            }
        }
        prepare(cycle, effector);
        break;
    case PUSH:    /* press(ActiveHand) */
        effector = NULL;
        for (effector1 = resources; effector1 != NULL; effector1 = effector1->next) {
            if (ColavitaIsHand(effector1)) {
                effector = most_salient_effector(cycle, effector, effector1);
            }
        }
        press(cycle, effector);
        break;
    case RELAX:    /* relax(ActiveHand) */
        effector = NULL;
        for (effector1 = resources; effector1 != NULL; effector1 = effector1->next) {
            if (ColavitaIsHand(effector1)) {
                effector = most_salient_effector(cycle, effector, effector1);
            }
        }
        relax(cycle, effector);
        break;
    default:
        fprintf(stderr, "WARNING: Unknown action: %d\n", schema->corresponding_action);
        break;
    }
}

BOOLEAN external_world(int c, TaskType task)
{
    extern BOOLEAN CoffeeStepWorld(int cycle);
    extern BOOLEAN SimpleVisualStepWorld(int cycle);
    extern BOOLEAN SimpleAuditoryStepWorld(int cycle);
    extern BOOLEAN ChoiceReactionTimeStepWorld(int cycle);
    extern BOOLEAN ConflictReactionTimeStepWorld(int cycle);

    switch(task) {
        case COFFEE_PREPARATION: case CEREAL_CONSUMPTION: case BREAKFAST: {
            return(CoffeeStepWorld(c));
        }
        case SIMPLE_VISUAL_RT: {
            return(SimpleVisualStepWorld(c));
        }
        case SIMPLE_AUDITORY_RT: {
            return(SimpleAuditoryStepWorld(c));
        }
        case CHOICE_RT: {
            return(ChoiceReactionTimeStepWorld(c));
        }
        case CONFLICT_RT: {
            return(ConflictReactionTimeStepWorld(c));
        }
        default: {
            fprintf(stderr, "Error: No world definition\n");
            return(TRUE);
        }
    }
}

/******************************************************************************/

